.TH audioemu 7D "27 Feb 2003"
.SH NAME
audioemu \- SoundBlaster Audigy audio device driver
.SH SYNOPSIS
.nf
.B audioemu:sound,audio
.B audioemu:sound,audioctl
.fi
.SH DESCRIPTION

The
.B audioemu
driver is a STREAMS based audio device driver for the SoundBlaster
Live! and SoundBlaster Audigy PCI family of audio cards. These cards
use an EMU10K1 or compatible chipset and include an AC'97 audio
codec and various other codecs (e.g. SPDIF in/out).

The Solaris
.B audioemu
driver was ported from NetBSD's SoundBlaster Live!
.B emuxki
driver and support for the EMU10K2/Audigy chipset was added.  It was
tested with an SoundBlaster Audigy (SB1394), Model SB0090 card
(EMU10K2 / Audigy Chipset); users have reported that the driver
works with the SoundBlaster Live! cards, too.
The
.B audioemu 
driver may or may not work with the SoundBlaster PCI512 (chipset
unknown, probably EMU10K1), and the
SoundBlaster Audigy 2 (chipset unknown, probably an enhanced
EMU10K2/Audigy chipset).

.I NOTE:
On Solaris SPARC, only the SoundBlaster Audigy is supported.
The DMA engine on the SoundBlaster Live! cannot access physical
addresses >=2GB, which is needed for SPARC DVMA.

.SH APPLICATION PROGRAMMING INTERFACE
The SoundBlaster Audigy device is treated as an exclusive resource; only
one process may open the device at a time for read or write access.
However, one process may open the device read-only (for recording
audio samples), while a second process may open the device write-only
(for playing audio samples).

This driver attempts to provide an implementation of the standard Sun audio
API.


.SS Audio Data Formats

The SoundBlaster Audigy handles 8-bit and 16-bit samples, in
mono or stereo mode.  Audio data is sampled at rates from 4,000
to 48,000 samples per seconds. 

The SoundBlaster Audigy handles 8-bit samples in excess-128 format
and 16-bit samples in two's complement format (16-bit PCM).  These two
formats correspond to an audio encoding of AUDIO_ENCODING_LINEAR8 with
a precision of 8 and an audio encoding of AUDIO_ENCODING_LINEAR with
precision 16. The
.B audioemu
driver also generates and accepts the encodings AUDIO_ENCODING_ULAW,
AUDIO_ENCODING_ALAW and AUDIO_ENCODING_LINEAR, all with precision 8.
The encodings AUDIO_ENCODING_ULAW and AUDIO_ENCODING_ALAW are
converted per software into 16-bit PCM format inside the driver.
Signed linear encoding with precision 8 is converted into 8-bit
excess-128 format.

.SS Audio Ports

There are three possible output ports available for the SoundBlaster
Audigy cards: AUDIO_SPEAKER, AUDIO_HEADPHONE and AUDIO_LINE_OUT.  Setting
either the AUDIO_SPEAKER or the AUDIO_HEADPHONE bit in the
.I play.port
field enables the card's speaker port. Setting the AUDIO_LINE_OUT bit in the
.I play.port
field selects the card's line level output port.  Output volume is
controlled by software.

The driver currently also duplicates the two analog front output
channels on the Audigy's analog rear output channels and the center /
subwoofer channel.  The volume on these duplicated outputs is not yet
controlled by software.

The
.I record.port
field of the audio information structure allows selection of which
audio source is used for recording, and may be set to one of
AUDIO_MICROPHONE, AUDIO_LINE_IN, AUDIO_CD, AUDIO_AUX1_IN,
AUDIO_AUX2_IN or AUDIO_CODEC_LOOPB_IN.  These select input from the
microphone jack, line-level input jack, internal CD input, internal
telephone answering device, internal audio from TV tuner or internal codec
loopback respectively.

.SH FILES
.PD 0
.TP 20
.B /dev/audio
linked to /dev/sound/0
.TP
.B /dev/audioctl
linked to /dev/sound/0ctl
.TP
.B /dev/sound/0
first audio device in the system
.TP
.B /dev/sound/0ctl
audio control for the first audio device
.PD
.SH SEE ALSO
.BR audioconvert (1),
.BR ioctl (2),
.BR audio (7I),
.BR streamio (7I)

.I ftp://download.intel.com/ial/scalableplatforms/audio/ac97r21.pdf
.LP
The EMU10K1 Digital Audio Processor, Thomas C. Savell, Joint E-mu/Creative
Technology Center, 1999
.LP
.I ftp://ftp.alsa-project.org/pub/datasheets/creative/m2049.pdf

.LP
FX8010 - A DSP Chip Architecture for Audio Effects, Steve Hoge, Joint
E-mu/Creative Technology Center
.LP
.I ftp://ftp.alsa-project.org/pub/datasheets/creative/HOG63.PS.gz

.SH NOTES
The
.B audioemu
driver will return "SUNW,CS4231" in the 
.I name
field of the audio_device structure.  This is the name of the audio
hardware in some SPARC based systems from Sun Microsystems and is not
the correct driver name for the SoundBlaster hardware.  The wrong
name was intentionally choosen to work around problems with hardcoded
audio_device names in various audio programs (audiotool, the Java JDK,
workman, ...) that derive information about the capabilities of the
audio device from the
.I name
field of the audio_device structure.  The name "SUNW,CS4231" was
choosen because this type of SPARC hardware supports roughly the same
features as the SoundBlaster Audigy PCI family of audio cards.

The
.I version
field of the audio_device structure contains the PCI revision byte of
the card as a two character hex string, a comma and the decoded
hardware name of the card. The
.I config
field contains "EMU10Kx".

.SH BUGS
The
.BR mixer (7I)
interface is not yet supported.

The SoundBlaster Audigy is able to play back from 64 channels (or 32
stereo voices) concurrently; the
.B audioemu
driver should not restrict playback to a single process.

Not all of the available audio input/output ports on the Audigy are
supported.  The driver currently supports the basic AC'97 input and
output ports (output: headphone, line-out; input: microphone, line-in,
analog CD).

During playback, there's a small amount of noise at the end of short clips.
