/*
    Soundblaster 16 PCI driver for Solaris.
    @(#) sbpci.c 2.1@(#)

    Written by Philip Brown <phil@bolthole.com>
    Rewritten by Jrgen Keil <jk@tools.de>

    Copyright  2000      Philip Brown
    Copyright  2000-2003 Jrgen Keil

    This driver is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This driver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this driver; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
*/

#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#if	SUNOS >= 509
#include <sys/audio/ac97.h>
#else
#include <sys/ac97.h>
#endif
#include <sys/audio.h>
#include "audioio_compat.h"
#include "ac97_codec.h"
#include "audio1371.h"

#include <sys/ddi.h>




static int audio1371_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int audio1371_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

static audio_info_t audio1371_defaults;
static ah_ad_ch_cap_t audio1371_ch_cap;
static ah_ad_cap_comb_t audio1371_cap_comb[];
static ah_ad_entry_t audio1371_entry;
static audio_device_t audio1371_dev_info;


static struct module_info audio1371_minfo = {
    0x5342,	/* module id number: "SB"; strlog(9F) module id, not used */
    "aud1371",	/* module name */
    0,		/* min packet size (0 for audio EOF packets) */
    16384,	/* max packet size accepted */
    48000*4*2,	/* ~2 secs of audio */
    0x8000
};


static struct qinit audio1371_rqueue = {
    audio_hlp_rput,
    audio_hlp_rsvc,
    audio_hlp_open,
    audio_hlp_close,
    NULL,
    &audio1371_minfo
};

static struct qinit audio1371_wqueue = {
    audio_hlp_wput,
    audio_hlp_wsvc,
    NULL,
    NULL,
    NULL,
    &audio1371_minfo
};


static struct streamtab audio1371_str_info = {
    &audio1371_rqueue,
    &audio1371_wqueue,
    NULL,
    NULL,
};

static struct cb_ops audio1371_cb_ops = {
    nulldev,			/* cb_open */
    nulldev,			/* cb_close */
    nodev,			/* cb_strategy */
    nodev,			/* cb_print */
    nodev,			/* cb_dump */
    nodev,			/* cb_read */
    nodev,			/* cb_write */
    nodev,			/* cb_ioctl */
    nodev,			/* cb_devmap */
    nodev,			/* cb_mmap */
    nodev,			/* cb_segmap */
    nochpoll,			/* cb_chpoll */
    ddi_prop_op,		/* cb_prop_op */
    &audio1371_str_info,	/* cb_str */
    D_NEW|D_MTSAFE|D_64BIT,	/* cb_flag */
    CB_REV,			/* cb_rev */
    nodev,			/* cb_aread */
    nodev,			/* cb_awrite */
};

static struct dev_ops  audio1371_drv_ops = {
    DEVO_REV,
    0,

    audio_hlp_getinfo,
    nulldev,			/* devo_identify */
    nulldev,			/* devo_probe */
    audio1371_attach,
    audio1371_detach,
    nodev,			/* devo_reset */

    &audio1371_cb_ops,
    NULL,			/* devo_bus_ops */
    NULL,			/* devo_power */
};

/*
 * Module linkage information for the kernel.
 */
static struct modldrv audio1371_modldrv = {
    &mod_driverops,		/* Type of module.  This one is a driver */
    "ensoniq 1371 audio driver",
    &audio1371_drv_ops		/* driver ops */
};


static struct modlinkage audio1371_modlinkage = {
    MODREV_1,
    { &audio1371_modldrv, NULL },
};


/*
 * This describes the card's DMA engine to the OS at hardware level
 * 
 * The ES1371 chip accesses memory in longword (4 bytes) quantities
 * (=> minxfer == 4) and has a 16-bit longword DMA counter (=> max DMA
 * xfer size == 0xffff * 4).  It does not support scatter/gatter DMA
 * transfers (=> s/g length == 1); sample data must occupy contiguous
 * physical memory.
 * 
 * The typical DMA transfer is a burst of 8 longwords (=> DMA burstsizes
 * has 0x20 bit set; I added 0x10, 0x8 and 0x4, too)
 */
static ddi_dma_attr_t dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0x0003fffc,			/* DMA counter register */
    1,				/* DMA address alignment */
    0x3c,			/* DMA burstsizes */
    4,				/* minxfer effective DMA size */
    0x0003ffff,			/* max DMA xfer size */
    0xffffffff,			/* segment boundary*/
    1,				/* s/g length */
				/*   (we want physically contigous mem) */
    4,				/* granularity of device */
				/*   (sbpci dma operates on longwords) */
    0				/* flags */
};


/* This describes things like byteorder internally for the pci card */
static struct ddi_device_acc_attr dev_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_STRUCTURE_LE_ACC,
    DDI_STRICTORDER_ACC
};


/* This describes things like byteorder internally for the pci card */
static struct ddi_device_acc_attr dev_sample_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_NEVERSWAP_ACC,
    DDI_STRICTORDER_ACC
};


uint_t audio1371_errlevel = AUD1371_LOG_L2;
uint_t audio1371_errmask =  AUD1371M_ALL;


static void
audio1371_vprintf(uint_t level, uint_t mask, char *fmt, va_list args)
{
    if (level <= audio1371_errlevel && (mask & audio1371_errmask))
	vcmn_err(CE_CONT, fmt, args);
}


#ifdef	DEBUG
void __attribute__ ((format (__printf__, 2, 3)))
audio1371_dprintf4(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audio1371_vprintf(AUD1371_LOG_L4, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audio1371_dprintf3(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audio1371_vprintf(AUD1371_LOG_L3, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audio1371_dprintf2(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audio1371_vprintf(AUD1371_LOG_L2, mask, fmt, args);
    va_end(args);
}
#endif

void __attribute__ ((format (__printf__, 2, 3)))
audio1371_dprintf1(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audio1371_vprintf(AUD1371_LOG_L1, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audio1371_dprintf0(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audio1371_vprintf(AUD1371_LOG_L0, mask, fmt, args);
    va_end(args);
}



static void *audio1371_state;

int 
_init(void)
{
    int err;

    err = ddi_soft_state_init(&audio1371_state,
			      sizeof(struct audio1371_softstate), 1);
    if (err == 0) {
	if ((err = mod_install(&audio1371_modlinkage)) != 0)
	    ddi_soft_state_fini(&audio1371_state);
    }

    DPRINTF4(AUD1371M_MODS, "_init, err=%d\n", err);

    return err;
}


int
_fini(void)
{
    int err;
    
    if ((err = mod_remove(&audio1371_modlinkage)) == 0)
	ddi_soft_state_fini(&audio1371_state);

    DPRINTF4(AUD1371M_MODS, "_fini, err=%d\n", err);

    return err;
}


int
_info(struct modinfo *modinfop)
{
    return mod_info(&audio1371_modlinkage, modinfop);
}


#define	PCI_VENDOR_ENSONIQ	0x1274
#define	PCI_PRODUCT_ENSONIQ_AUDIOPCI	0x5000
#define	PCI_PRODUCT_ENSONIQ_AUDIOPCI97	0x1371
#define	PCI_PRODUCT_ENSONIQ_CT5880	0x5880


/* returns AUDIO1371_FAILURE/AUDIO1371_SUCCESS */
static int
audio1371_pci_setup(struct audio1371_softstate *softstate)
{
    dev_info_t *dip = softstate->dip;
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;
    uint16_t pci_status;
    uint16_t pci_vendorid;
    uint16_t pci_deviceid;
    uint8_t  pci_revision;
    uint8_t  pci_intrinfo;
    uint16_t pci_commandreg;
    uint8_t  pci_lattimerreg;
    char *revstr = "";

    pci_status   = pci_config_get16(conf_handle, PCI_CONF_STAT);
    pci_vendorid = pci_config_get16(conf_handle, PCI_CONF_VENID);
    pci_deviceid = pci_config_get16(conf_handle, PCI_CONF_DEVID);
    pci_revision = pci_config_get8 (conf_handle, PCI_CONF_REVID);
    pci_intrinfo = pci_config_get8 (conf_handle, PCI_CONF_ILINE);

    softstate->ens_chip_type =
	(pci_vendorid == PCI_VENDOR_ENSONIQ &&
	 pci_deviceid == PCI_PRODUCT_ENSONIQ_AUDIOPCI)
	? ENS_CHIP_1370
	: ENS_CHIP_1371;
    softstate->ct5880 = 0;

    switch (softstate->ens_chip_type) {
    case ENS_CHIP_1370:
	softstate->ens_chip_ops = &ens1370_ops;
	/* On ENS1370: Only DAC2 supports variable PLL clock rates */
	softstate->playback_dac = 2;
	revstr = "ES1370";
	break;

    case ENS_CHIP_1371:
	softstate->ens_chip_ops = &ens1371_ops;
	/* 
	 * Both DACs support variable sample rates, but for some reason
	 * DAC1 playback can be stopped immediately, so we prefer DAC1
	 * on the ENS1371
	 */
	softstate->playback_dac = 1;
	if (pci_vendorid == PCI_VENDOR_ENSONIQ &&
	    pci_deviceid == PCI_PRODUCT_ENSONIQ_CT5880)
	    softstate->ct5880 = 1;
	switch (pci_revision) {
	case SBPCI_EV1938_A: revstr = "EV1938A"; break;
	case SBPCI_CT5880_C: revstr = "CT5880C"; softstate->ct5880 = 1; break;
	case SBPCI_ES1373_A: revstr = "ES1373A"; break;
	case SBPCI_ES1373_B: revstr = "ES1373B"; break;
	case SBPCI_CT5880_A: revstr = "CT5880A"; softstate->ct5880 = 1; break;
	case SBPCI_ES1371_B: revstr = "ES1371B"; break;
	case SBPCI_ES1373_8: revstr = "ES1373";  break;
	}
	break;
    }
    softstate->rev = pci_revision;
    softstate->revstr = revstr;

    /* Announce the hw we're attaching to */
    ddi_report_dev(dip);
    cmn_err(CE_CONT, "?PCI-device: %s, audio1371%d, %s%s(rev. 0x%02x) at irq %d\n",
	    ddi_node_name(dip),
	    ddi_get_instance(dip), 
	    revstr, revstr[0] ? " " : "", pci_revision,
	    pci_intrinfo);

#if i386
    /*
     * reject interrupt line 255, per PCI spec 6.2.4, footnote 43,
     * this means "unknown" or "not connected" to the interrupt
     * controller on x86 based PCs.
     * Sun drivers (iprb) reject iline 0 as well.
     */
    if (pci_intrinfo == 0 || pci_intrinfo == 255) {
	cmn_err(CE_WARN, "interrupt line value out of range: %d", pci_intrinfo);
	return AUDIO1371_FAILURE;
    }
#endif
    
    pci_commandreg = pci_config_get16(conf_handle, PCI_CONF_COMM);
    DPRINTF3(AUD1371M_ATTACH,
	     "pci command register 0x%x %s%s%s%s",
	     pci_commandreg,
	     pci_commandreg & PCI_COMM_SERR_ENABLE ? " serr#_en"  : "",
	     pci_commandreg & PCI_COMM_ME  ? " busmaster" : "",
	     pci_commandreg & PCI_COMM_MAE ? " mem_access" : "",
	     pci_commandreg & PCI_COMM_IO  ? " io_access" : "");
    if ((pci_commandreg & (PCI_COMM_IO|PCI_COMM_ME)) == (PCI_COMM_IO|PCI_COMM_ME)) {
	DPRINTF3(AUD1371M_ATTACH, "busmastering and io_access already enabled");
    } else {
	DPRINTF3(AUD1371M_ATTACH, "Enabling PCI I/O and bus mastering access");
	pci_config_put16(conf_handle, PCI_CONF_COMM, pci_commandreg | PCI_COMM_IO | PCI_COMM_ME);
    }

#if	DIAGNOSTICS
    /* 
     * Clear master/target abort status, if set.  Note that bits in the pci
     * status register are cleared by writing a 1 bit into that bit
     * position!
     */
    if ((pci_status &= (PCI_STAT_R_MAST_AB|PCI_STAT_R_TARG_AB)) != 0) {
	pci_config_put16(conf_handle, PCI_CONF_STAT, pci_status);
	DPRINTF3(AUD1371M_ATTACH, "cleared pci status bits: %x\n", pci_status);
    }
#endif

    /*
     * (linux alsa drivers) increase the pci latency timer (from 32 -> 64).
     * This fixes the "received master abort" status we're seeing for
     * each DMA burst access.
     */
    pci_lattimerreg = pci_config_get8(conf_handle, PCI_CONF_LATENCY_TIMER);
    if (pci_lattimerreg < 64) {
	DPRINTF3(AUD1371M_ATTACH, "PCI latency timer: increase from %d to 64",
		pci_lattimerreg);
	pci_lattimerreg = 64;
	pci_config_put8(conf_handle, PCI_CONF_LATENCY_TIMER,
			pci_lattimerreg);
    } else {
	DPRINTF3(AUD1371M_ATTACH, "PCI latency timer %d", pci_lattimerreg);
    }

    return AUDIO1371_SUCCESS;
}


#if	DIAGNOSTICS
static void
check_pci_status(struct audio1371_softstate *softstate, char *where)
{
    uint16_t pci_status;

    if (softstate->pci_conf_handle == NULL)
	return;

    pci_status = pci_config_get16(softstate->pci_conf_handle, PCI_CONF_STAT);
    if (pci_status & (PCI_STAT_R_MAST_AB|PCI_STAT_R_TARG_AB)) {
	printf("%s, intr# %d, pci_status = %04x",
	       where, softstate->intr_count, pci_status);
	pci_status &= (PCI_STAT_R_MAST_AB|PCI_STAT_R_TARG_AB);
	pci_config_put16(softstate->pci_conf_handle, PCI_CONF_STAT, pci_status);
    }
}
#endif


/*
 * allocate a sample buffer, in DMAable memory
 * for use by the playback/record hardware
 */
static int
audio1371_alloc_buffer(dev_info_t *dip, sample_buffer_t *buffer)
{
    buffer->dma_bound = FALSE;

    /* ring buffer is empty and idle */
    buffer->offs = 0;
    buffer->used = 0;
    buffer->state = SAMPLE_BUF_IDLE;

    if (ddi_dma_alloc_handle(dip, &dma_attr,
			     DDI_DMA_DONTWAIT, NULL,
			     &buffer->dma_handle) != DDI_SUCCESS) {
	/*
	 * make sure that we wont free any junk, in case ddi_dma_alloc_handle
	 * has put something into buffer->dma_handle on an allocation failure
	 */
	buffer->dma_handle = 0;
	return AUDIO1371_FAILURE;
    }

    if (ddi_dma_mem_alloc(buffer->dma_handle,
			  SAMPLEBUF_SIZE,
			  &dev_sample_attr,
			  DDI_DMA_STREAMING,
			  DDI_DMA_DONTWAIT, NULL,
			  &buffer->data,
			  &buffer->reallen,
			  &buffer->acc_handle) != DDI_SUCCESS) {
	return AUDIO1371_FAILURE;
    }

    return AUDIO1371_SUCCESS;
}


/*
 * free sample buffer resources, undo allocation from audio1371_alloc_buffer
 */
static int
audio1371_free_buffer(sample_buffer_t *buffer)
{
    if (buffer->dma_handle && buffer->dma_bound) {
	ddi_dma_unbind_handle(buffer->dma_handle);
	buffer->dma_bound = FALSE;
    }

    if (buffer->acc_handle != NULL)
	ddi_dma_mem_free(&buffer->acc_handle);

    if (buffer->dma_handle)
	ddi_dma_free_handle(&buffer->dma_handle);

    return AUDIO1371_SUCCESS;
}


static int
audio1371_cleanup(struct audio1371_softstate *softstate)
{
    if (softstate->audio_handle) {
	if (audio_hlp_detach(softstate->audio_handle, DDI_DETACH) != DDI_SUCCESS)
	    return DDI_FAILURE;
    }

    if (softstate->codec_handle) {
	CODEC_DELETE(softstate->codec_handle);
	softstate->codec_handle = NULL;
    }

    /*
     * remove interrupt handler before we free the card's mapped registers
     * otherwise an interrupt could access softstate->regs_handle in our
     * interrupt handler
     */
    if (softstate->intr_cookie != NULL) {
	ddi_remove_intr(softstate->dip, 0, softstate->intr_cookie);
    }

    if (softstate->regs_handle != NULL) {
	ddi_regs_map_free(&softstate->regs_handle);
    }

    if (softstate->pci_conf_handle != NULL) {
	pci_config_teardown(&softstate->pci_conf_handle);
    }

    audio1371_free_buffer(&softstate->play_buf);
    audio1371_free_buffer(&softstate->record_buf);

    mutex_destroy(&softstate->instance_mutex);
    mutex_destroy(&softstate->intr_mutex);

    ddi_soft_state_free(audio1371_state, ddi_get_instance(softstate->dip));

    return DDI_SUCCESS;
}


/*
 * When accessing "memory1","memory2" stuff for FRAME window settings, you
 * first need to make sure you're on the right "page".
 * page 0xc is DAC1/DAC2
 * page 0xd is the ADC stuff.
 */
static void
select_mempage(struct audio1371_softstate *softstate, int pagenum)
{
    SBPCI_PUT32(SBPCI_MEMPAGE, pagenum & 0x0f);
}


#if	0
static void
print_pcistatus(dev_info_t *dip)
{
    ddi_acc_handle_t conf_handle;
    uint16_t pcistatusreg;

    if ((conf_handle = softstate->pci_conf_handle) == NULL)
	return;

    pcistatusreg = pci_config_get16(conf_handle, PCI_CONF_STAT);
    cmn_err(CE_CONT, "pci status reg: %x", pcistatusreg);
    if (pcistatusreg & PCI_STAT_PERROR) {
	cmn_err(CE_CONT, "pci_status: parity error detected");
    }
    if (pcistatusreg & PCI_STAT_S_SYSERR) {
	cmn_err(CE_CONT, "pci_status: signaled system error");
    }
    if (pcistatusreg & PCI_STAT_R_MAST_AB) {
	cmn_err(CE_CONT, "pci_status: received master-abort");
    }
    if (pcistatusreg & PCI_STAT_R_TARG_AB) {
	cmn_err(CE_CONT, "pci_status: received target-abort");
    }
    if (pcistatusreg & PCI_STAT_S_TARG_AB) {
	cmn_err(CE_CONT, "pci_status: signaled target-abort");
    }

#ifdef	PCI_STAT_CAP
    if (pcistatusreg & PCI_STAT_CAP) {
	uint16_t pcipmcapreg;
	uint16_t pcipmstatreg;
	uint8_t  capptr, capid, capnext;
	/*
	 * we have a PCI capabilities list (linked list in pci
	 * configuration space 
	 */
	cmn_err(CE_CONT, "pci_status: capabilties list");
	capptr = pci_config_get8(conf_handle, PCI_CONF_CAP_PTR);
	cmn_err(CE_CONT, "pci capabilities ptr: %x", capptr);
	for (; (capptr &= ~3) != PCI_CAP_NEXT_PTR_NULL; capptr = capnext) {
	    capid   = pci_config_get8(conf_handle, capptr + PCI_CAP_ID);
	    capnext = pci_config_get8(conf_handle, capptr + PCI_CAP_NEXT_PTR);
	    cmn_err(CE_CONT, "pci capabilities id: %x", capid);
	    cmn_err(CE_CONT, "pci capabilities link: %x", capnext);

	    switch (capid) {
	    case PCI_CAP_ID_PM:
		pcipmcapreg = pci_config_get16(conf_handle, capptr + PCI_PMCAP);
		cmn_err(CE_CONT, "pci power management capabilities reg: %x", pcipmcapreg);
		if (pcipmcapreg & PCI_PMCAP_D2) cmn_err(CE_CONT, "D2 is supported");
		if (pcipmcapreg & PCI_PMCAP_D1) cmn_err(CE_CONT, "D1 is supported");
		cmn_err(CE_CONT, "PCI power managment v%d supported", pcipmcapreg & 7);

		pcipmstatreg = pci_config_get16(conf_handle, capptr + PCI_PMCSR);
		cmn_err(CE_CONT, "pci power management status reg: %x", pcipmstatreg);
		cmn_err(CE_CONT, "pci power management state: D%d", pcipmstatreg & 3);
		break;
	    }
	}
    }
#endif

}
#endif


/************************************************************/
/*    End of Codec section. Start of SampleRate section     */
/************************************************************/


/*
 * Stop the device, prepare for unloading the driver.
 */
static int
audio1371_stop_device(struct audio1371_softstate *softstate)
{
    SBPCI_PUT32(SBPCI_CONTROL, 0);
    SBPCI_PUT32(SBPCI_SERIAL,  0);
    SBPCI_PUT32(SBPCI_LEGACY,  0);

    SBPCI_PUT16(SBPCI_DAC1_COUNT, 0);
    SBPCI_PUT16(SBPCI_DAC2_COUNT, 0);
    SBPCI_PUT16(SBPCI_ADC_COUNT,  0);

    /* disable all DAC/ADC sample rate conversion */
    ENS_ENABLE_SAMPLE_RATE_CONV(softstate, 0, FALSE, NULL);
    ENS_ENABLE_SAMPLE_RATE_CONV(softstate, 1, FALSE, NULL);
    ENS_ENABLE_SAMPLE_RATE_CONV(softstate, 2, FALSE, NULL);
    ENS_ENABLE_SAMPLE_RATE_CONV(softstate, 3, FALSE, NULL);

    return AUDIO1371_SUCCESS;
}


/*
 * This attempts to reset the ens1371-compatible card
 */
static int
audio1371_reset_card(struct audio1371_softstate *softstate,
		     audio_info_t		*audio_info,
		     int			 verbose)
{
    int count;

    SBPCI_PUT32(SBPCI_SERIAL,  0);
    SBPCI_PUT32(SBPCI_CONTROL, 0);
    SBPCI_PUT32(SBPCI_LEGACY,  0);

    SBPCI_PUT16(SBPCI_DAC1_COUNT, 0);
    SBPCI_PUT16(SBPCI_DAC2_COUNT, 0);
    SBPCI_PUT16(SBPCI_ADC_COUNT,  0);

    DPRINTF4(AUD1371M_CODEC, "calling reset_codec from reset_card()");
    ENS_CODEC_INTERFACE_RESET(softstate);
    /*printf("reset card, verbose = %d\n", verbose);*/

    DPRINTF4(AUD1371M_CODEC, "calling configure_pm D0 from reset_card()");
    ENS_CONFIGURE_PM(softstate, 0);

    /*
     * clear sound cache ram
     */
    for (count = 0; count < 16; count++) {
	select_mempage(softstate, count);

	SBPCI_PUT32(SBPCI_MEMORY,    0);
	SBPCI_PUT32(SBPCI_MEMORY+4,  0);
	SBPCI_PUT32(SBPCI_MEMORY+8,  0);
	SBPCI_PUT32(SBPCI_MEMORY+12, 0);
    }

    /*
     * disable "sample rate" hardware, program it, and then
     * fully enable it. 
     *
     * If the "sample rate converter" is not properly initialized,
     * then turning this on will lock up the chip.
     * Power-flip time.
     */
    ENS_INIT_SAMPLE_RATE_CONV(softstate);

    ENS_SETADCFREQ(softstate, audio_info->record.sample_rate);
    ENS_SETDACFREQ(softstate, audio_info->play.sample_rate, 1);
    ENS_SETDACFREQ(softstate, audio_info->play.sample_rate, 2);

    /* "global" sample rate converter enable */
    ENS_ENABLE_SAMPLE_RATE_CONV(softstate, 0, TRUE, NULL);

    if (!CODEC_RESET(softstate->codec_handle, audio_info, verbose)) {
	DPRINTF2(AUD1371M_CODEC, "Audio CODEC reset failed");
	return AUDIO1371_FAILURE;
    }

    return AUDIO1371_SUCCESS;
}


/*
 * Setup the card's DMA hardware to access the playback ring buffer
 */
static int
audio1371_setup_dac_dma(struct audio1371_softstate *softstate,
			int DACnum,
			sample_buffer_t *buffer)
{
    uint_t curr_count, len;
    int bindstat;
    caddr_t vaddr;
    ddi_dma_cookie_t dma_cookie;
    uint_t ccount;

    bindstat = ddi_dma_addr_bind_handle(
		buffer->dma_handle,
		NULL,
		buffer->data, buffer->reallen,
		DDI_DMA_WRITE | DDI_DMA_STREAMING | DDI_DMA_REDZONE,
		DDI_DMA_DONTWAIT, NULL,
		&dma_cookie, &ccount);
    vaddr = buffer->data;

    if (bindstat != DDI_DMA_MAPPED) {
	DPRINTF4(AUD1371M_INTR, "Could not map DMA chunk at %p, bindstat=%d",
		vaddr, bindstat);
	if (bindstat == DDI_DMA_PARTIAL_MAP)
	    goto unbind_and_error;

	/* else failure, do not unbind dma handle */
	goto error;
    }

    buffer->dma_bound = TRUE;

    if (ccount > 1) {
	/* 
 	 * should not happen, because we've allocated physically
	 * contiguous ram. (The SB PCI card does not support
	 * scatter-gatter type DMA,  so a cookie count > 1
	 * condition is rejected)
	 */
	DPRINTF4(AUD1371M_INTR, "More than one cookie (%d)", ccount);
	for (len = 0; len < ccount; len++) {
	    if (len > 0)
		ddi_dma_nextcookie(buffer->dma_handle, &dma_cookie);
	    DPRINTF4(AUD1371M_INTR, "cookie #%d: paddr 0x%p, len 0x%lx",
		    len,
		    (caddr_t)(long)dma_cookie.dmac_address,
		    (long)dma_cookie.dmac_size);
	}
	DPRINTF4(AUD1371M_INTR, "  Unbinding handle, failing playback");
	goto unbind_and_error;
    }

    curr_count = 0;
    len = SAMPLEBUF_SIZE;
    DPRINTF4(AUD1371M_INTR, "setup_dac_dma: virt=0x%p, phys=0x%p",
	    vaddr, (caddr_t)(long)dma_cookie.dmac_address);


    /* setup the DMA engine */
    select_mempage(softstate, SBPCI_DAC_MEMPG);

    if (DACnum == 1) {

	SBPCI_PUT32(SBPCI_DAC1_FRAME1, dma_cookie.dmac_address);
	/* Size of ring buffer in "longwords" (lower 16 bits) */
	SBPCI_PUT32(SBPCI_DAC1_FRAME2, (curr_count << 16) | ((len >> 2) - 1));

    } else {

	SBPCI_PUT32(SBPCI_DAC2_FRAME1, dma_cookie.dmac_address);
	/* Size of ring buffer in "longwords" (lower 16 bits) */
	SBPCI_PUT32(SBPCI_DAC2_FRAME2, (curr_count << 16) | ((len >> 2) - 1));
    }

    return AUDIO1371_SUCCESS;


 unbind_and_error:
    ddi_dma_unbind_handle(buffer->dma_handle);
 error:
    buffer->dma_bound = FALSE;
    return AUDIO1371_FAILURE;
}


/*
 * Convert a number of bytes into a number of samples, according to
 * a given audio format 'prinfo'.
 */
static int
audio1371_bytes_to_samples(int byte_count, audio_prinfo_t *prinfo)
{
    int pcm16, stereo, sample_count;

    pcm16 = (prinfo->precision == AUDIO_PRECISION_16 ||
	     (prinfo->encoding != AUDIO_ENCODING_LINEAR &&
	      prinfo->encoding != AUDIO_ENCODING_LINEAR8));
    stereo = (prinfo->channels == AUDIO_CHANNELS_STEREO);

    sample_count = byte_count;
    if (pcm16)  sample_count >>= 1;
    if (stereo) sample_count >>= 1;
    return sample_count;
}


/*
 * Convert a number of bytes into a number of samples, according to
 * a given audio format 'prinfo'.
 */
static int
audio1371_samples_to_bytes(int sample_count, audio_prinfo_t *prinfo)
{
    int pcm16, stereo, byte_count;

    pcm16 = (prinfo->precision == AUDIO_PRECISION_16 ||
	     (prinfo->encoding != AUDIO_ENCODING_LINEAR &&
	      prinfo->encoding != AUDIO_ENCODING_LINEAR8));
    stereo = (prinfo->channels == AUDIO_CHANNELS_STEREO);

    byte_count = sample_count;
    if (pcm16)  byte_count <<= 1;
    if (stereo) byte_count <<= 1;
    return byte_count;
}


/*
 * Setup the DAC sample count register to arrange for a device
 * interrupt after a certain amount of samples have been played.
 */
static void
audio1371_set_dac_samplecount(struct audio1371_softstate *softstate,
			      int DACnum,
			      sample_buffer_t *buffer)
{
    int byte_count;
    int sample_count;
    uint32_t serialval;

    /* 
     * interrupt after the current contents of the ring buffer has
     * been played,  but make sure that we get an interrupt every
     * SAMPLEBUF_INTR_CHUNK bytes (currently every 8 kbyte).
     */
    byte_count = buffer->used;
    if (byte_count > SAMPLEBUF_INTR_CHUNK)
	byte_count = SAMPLEBUF_INTR_CHUNK;
    sample_count = audio1371_bytes_to_samples(byte_count,
					      &softstate->audio_info.play);

    serialval = SBPCI_GET32(SBPCI_SERIAL);
    if (DACnum == 1) {
	SBPCI_PUT16(SBPCI_DAC1_COUNT, sample_count - 1);
	/* notify hw of sample counter change */
        SBPCI_PUT32(SBPCI_SERIAL,
		    (serialval & ~SBPCI_SERIAL_ENABLEDAC1) 
		    		| SBPCI_SERIAL_SCT_RLD_DAC1);
	drv_usecwait(1);
	/* clear sample counter reload bit, enable interrupts */
        SBPCI_PUT32(SBPCI_SERIAL,
		    (serialval & ~SBPCI_SERIAL_SCT_RLD_DAC1)
		    		| SBPCI_SERIAL_ENABLEDAC1);
    } else {
	SBPCI_PUT16(SBPCI_DAC2_COUNT, sample_count - 1);
	/*
	 * dac2 has no "sample counter reload" notification,
	 * just enable interrupts
	 */
        SBPCI_PUT32(SBPCI_SERIAL, serialval | SBPCI_SERIAL_ENABLEDAC2);
    }

    DPRINTF4(AUD1371M_INTR, "set dac samples: %x (total avail %x)",
	    sample_count - 1, buffer->used);
}


static void
audio1371_configure_playback(struct audio1371_softstate *softstate, int DACnum)
{
    uint32_t serialval;
    int pcm16;
    int stereo;
    audio_prinfo_t *prinfo = &softstate->audio_info.play;

    DPRINTF4(AUD1371M_INTR, "configure_playback: %dhz, %d-bit, %s, %d-channel",
	    prinfo->sample_rate,
	    prinfo->precision,
	    (prinfo->encoding == AUDIO_ENCODING_ULAW
	     ? "uLaw"
	     : (prinfo->encoding == AUDIO_ENCODING_LINEAR
		? "linear"
		: "???")),
	    prinfo->channels);

    /* setup DAC playback format */
    serialval = SBPCI_GET32(SBPCI_SERIAL);
    serialval &= ~SBPCI_SERIAL_P2_ST_INC_MASK;
    serialval &= ~SBPCI_SERIAL_P2_END_INC_MASK;
    pcm16 = (prinfo->precision == AUDIO_PRECISION_16 ||
	     (prinfo->encoding != AUDIO_ENCODING_LINEAR &&
	      prinfo->encoding != AUDIO_ENCODING_LINEAR8));
    stereo = (prinfo->channels == AUDIO_CHANNELS_STEREO);
    serialval |= SBPCI_SERIAL_P2_ST_INC(0);
    serialval |= SBPCI_SERIAL_P2_END_INC(pcm16 ? 2 : 1);
    if (DACnum == 1) {
	serialval &= ~(SBPCI_SERIAL_DAC1W16 | SBPCI_SERIAL_DAC1STEREO);
	if (stereo)
	    serialval |= SBPCI_SERIAL_DAC1STEREO;
	if (pcm16)
	    serialval |= SBPCI_SERIAL_DAC1W16;
	serialval &= ~SBPCI_SERIAL_NOLOOPDAC1;
	serialval &= ~SBPCI_SERIAL_PAUSEDAC1;
    } else {
	serialval &= ~(SBPCI_SERIAL_DAC2W16 | SBPCI_SERIAL_DAC2STEREO);
	if (stereo)
	    serialval |= SBPCI_SERIAL_DAC2STEREO;
	if (pcm16)
	    serialval |= SBPCI_SERIAL_DAC2W16;

	/* 
	 * play back the last sample while DAC2 is stopped
	 * (e.g.: during DMA reload)
	 */
	serialval |= SBPCI_SERIAL_DAC2_SEN;
	serialval &= ~SBPCI_SERIAL_NOLOOPDAC2;
	serialval &= ~SBPCI_SERIAL_PAUSEDAC2;
    }
    SBPCI_PUT32(SBPCI_SERIAL, serialval);

    ENS_SETDACFREQ(softstate, prinfo->sample_rate, DACnum);
}


/*
 * Start playback for the samples in play_buf
 */
static void
audio1371_play_start_unlocked(struct audio1371_softstate *softstate, int DACnum)
{
    uint32_t controlval;
    uint32_t dac_enable_bit;
    int loop;
    int dac_count_reg;
    uint16_t curr_samp_ct;
    
    if (DACnum != 1 && DACnum != 2) {
	DPRINTF4(AUD1371M_INTR, "audio1371_play_start: Sorry, we only do DAC1/2 right now");
	return;
    }

    dac_enable_bit = (DACnum == 1 ? SBPCI_CTRL_DAC1 : SBPCI_CTRL_DAC2);

    /* disable DAC1/2 playback channel */
    controlval = SBPCI_GET32(SBPCI_CONTROL);
    SBPCI_PUT32(SBPCI_CONTROL, controlval & ~dac_enable_bit);

    audio1371_configure_playback(softstate, DACnum);

    /* enable the sample rate converter for the DAC's playback channel */
    ENS_ENABLE_SAMPLE_RATE_CONV(softstate, DACnum, TRUE, NULL);

    if (audio1371_setup_dac_dma(softstate, DACnum, &softstate->play_buf) 
				== AUDIO1371_SUCCESS) {

#if 0
	softstate->audio_info.play.active = 1;
#endif

	audio1371_set_dac_samplecount(softstate, DACnum, &softstate->play_buf);

	/* enable the DAC1/2 playback channel */
	SBPCI_PUT32(SBPCI_CONTROL, controlval | dac_enable_bit);

	/*
	 * NOTE: the following delay is intended for a correct
	 * audio.play.samples counter in the AUDIO_GETINFO ioctl.
	 *
	 * Previously, there were sporadic cases of a decreasing
	 * audio.play.samples counter, because an audio AUDIO_GETINFO
	 * immediately after playback start would still see the
	 * current sample counter from the last playback - the HW has not
	 * yet started playing samples.
	 *
	 * After the DCA is enabled, there's a small delay until the
	 * hardware updates the current sample counter.
	 */
	dac_count_reg = DACnum == 1 ? SBPCI_DAC1_COUNT : SBPCI_DAC2_COUNT;

	curr_samp_ct = SBPCI_GET32(dac_count_reg) >> 16;
	for (loop = 0; loop < 50; loop++) {
	    drv_usecwait(10);
	    if (SBPCI_GET32(dac_count_reg) >> 16 != curr_samp_ct)
		break;
	}
	DPRINTF4(AUD1371M_INTR, "waited %d x 10usec for playback start", loop);
    }
}


/*
 * Stop playback of samples
 */
static void
audio1371_play_stop_unlocked(struct audio1371_softstate *softstate, int DACnum)
{
    uint32_t control;

    /*
     * XXX (what a mess):
     * It seems that disabling DAC1 immediately stops audio playback.
     * For some reason this is not true for DAC2; somehow it continues
     * to play "noise" after it's disabled.
     *
     * Unfortunatelly it's DAC2 that supports variable samples rates
     * on an ENS1370 card via the PLL clock; DAC1 on an ENS1370 only
     * supports four fixed frequencies.
     */
    if (softstate->ens_chip_type != ENS_CHIP_1370) {
	/*
	 * XXX: on ens1370: noise at end of playback if the audio
	 * device is still open via the "ctl" device,
	 * e.g. audiocontrol.  That is, when AD_TEARDOWN is not called.
	 *
	 * On the ens1371, this code works well.
	 */
	control = SBPCI_GET32(SBPCI_CONTROL);
	if (DACnum == 1) {
	    SBPCI_PUT32(SBPCI_CONTROL, control & ~SBPCI_CTRL_DAC1);
	} else {
	    SBPCI_PUT32(SBPCI_CONTROL, control & ~SBPCI_CTRL_DAC2);
	}
	ENS_ENABLE_SAMPLE_RATE_CONV(softstate, DACnum, FALSE, NULL);
    } else {
	/*
	 * XXX: ens1371 hw does not intr after playing 11025/mono/U8
	 *
	 * I've not yet observed this problem on the ens1370.
	 */
	uint32_t serial;

	control = SBPCI_GET32(SBPCI_CONTROL);
	serial  = SBPCI_GET32(SBPCI_SERIAL);
	if (DACnum == 1) {
	    serial |= SBPCI_SERIAL_NOLOOPDAC1;
	    serial |= SBPCI_SERIAL_PAUSEDAC1;
	    serial &= ~SBPCI_SERIAL_ENABLEDAC1;
	    SBPCI_PUT32(SBPCI_SERIAL,  serial);

	    SBPCI_PUT32(SBPCI_CONTROL, control & ~SBPCI_CTRL_DAC1);

	    SBPCI_PUT16(SBPCI_DAC1_COUNT, 0);
	} else {
	    serial |= SBPCI_SERIAL_NOLOOPDAC2;
#if 0
	    serial &= ~SBPCI_SERIAL_DAC2_SEN;	/* play zero when stopped */
	    serial |= SBPCI_SERIAL_DAC2_SEN;	/* play last samp when stopped */
#endif
	    serial |= SBPCI_SERIAL_PAUSEDAC2;
	    serial &= ~SBPCI_SERIAL_ENABLEDAC2;
	    SBPCI_PUT32(SBPCI_SERIAL, serial);
	    
	    SBPCI_PUT32(SBPCI_CONTROL, control & ~SBPCI_CTRL_DAC2);

	    SBPCI_PUT16(SBPCI_DAC2_COUNT, 0);
	}

	ENS_ENABLE_SAMPLE_RATE_CONV(softstate, DACnum, FALSE, NULL);

	bzero(softstate->play_buf.data, SAMPLEBUF_SIZE);
    }

    if (softstate->play_buf.dma_handle && softstate->play_buf.dma_bound) {
	ddi_dma_unbind_handle(softstate->play_buf.dma_handle);
	softstate->play_buf.dma_bound = FALSE;
    }

    softstate->play_buf.offs = 0;
    softstate->play_buf.used = 0;
    softstate->play_buf.state = SAMPLE_BUF_IDLE;
}


static void
audio1371_play_stop(struct audio1371_softstate *softstate, int DACnum)
{
    if (softstate->play_timeout_id) {
	untimeout(softstate->play_timeout_id);
	softstate->play_timeout_id = 0;
    }

    mutex_enter(&softstate->intr_mutex);
    audio1371_play_stop_unlocked(softstate, DACnum);
    mutex_exit(&softstate->intr_mutex);
}


static void
audio1371_play_pause(struct audio1371_softstate *softstate, int DACnum, int stop)
{
    uint32_t serial;
    uint32_t pause_bit;

    pause_bit = (DACnum == 1
		 ? SBPCI_SERIAL_PAUSEDAC1
		 : SBPCI_SERIAL_PAUSEDAC2);

    mutex_enter(&softstate->intr_mutex);
    serial = SBPCI_GET32(SBPCI_SERIAL);
    if (stop)
	serial |= pause_bit;
    else
	serial &= ~pause_bit;
    SBPCI_PUT32(SBPCI_SERIAL, serial);
    mutex_exit(&softstate->intr_mutex);
}


/*
 * Program the card's DMA hardware to save recorded data into the
 * record ring buffer
 */
static int
audio1371_setup_adc_dma(struct audio1371_softstate *softstate,
			sample_buffer_t *buffer)
{
    uint_t curr_count, len;
    int bindstat;
    caddr_t vaddr;
    ddi_dma_cookie_t dma_cookie;
    uint_t ccount;

    bindstat = ddi_dma_addr_bind_handle(
		buffer->dma_handle,
		NULL,
		buffer->data, buffer->reallen,
		DDI_DMA_READ | DDI_DMA_STREAMING | DDI_DMA_REDZONE,
		DDI_DMA_DONTWAIT, NULL,
		&dma_cookie, &ccount);
    vaddr = buffer->data;

    if (bindstat != DDI_DMA_MAPPED) {
	DPRINTF4(AUD1371M_INTR, "Could not map DMA chunk at %p, bindstat=%d",
		vaddr, bindstat);
	if (bindstat == DDI_DMA_PARTIAL_MAP)
	    goto unbind_and_error;

	/* else failure, do not unbind dma handle */
	goto error;
    }

    buffer->dma_bound = TRUE;

    if (ccount > 1) {
	/* 
 	 * should not happen, because we've allocated physically
	 * contiguous ram
	 */
	DPRINTF4(AUD1371M_INTR, "More than one cookie (%d)", ccount);
	for (len = 0; len < ccount; len++) {
	    if (len > 0)
		ddi_dma_nextcookie(buffer->dma_handle, &dma_cookie);
	    DPRINTF4(AUD1371M_INTR, "cookie #%d: paddr 0x%p, len 0x%lx",
		    len,
		    (caddr_t)(long)dma_cookie.dmac_address,
		    (long)dma_cookie.dmac_size);
	}
	DPRINTF4(AUD1371M_INTR, "  Unbinding handle, failing recording");
	goto unbind_and_error;
    }

    curr_count = 0;
    len = SAMPLEBUF_SIZE;
    DPRINTF4(AUD1371M_INTR, "setup_adc_dma: virt=0x%p, phys=0x%p",
	    vaddr, (caddr_t)(long)dma_cookie.dmac_address);


    /* setup the DMA engine */
    select_mempage(softstate, SBPCI_ADC_MEMPG);

    SBPCI_PUT32(SBPCI_ADC_FRAME1, dma_cookie.dmac_address);
    /* Size of mapped chunk in "longwords" */
    SBPCI_PUT32(SBPCI_ADC_FRAME2, (curr_count << 16) | ((len >> 2) - 1));

    return AUDIO1371_SUCCESS;


 unbind_and_error:
    ddi_dma_unbind_handle(buffer->dma_handle);
 error:
    buffer->dma_bound = FALSE;

    return AUDIO1371_FAILURE;
}


static void
audio1371_configure_recording(struct audio1371_softstate *softstate)
{
    uint32_t serialval;
    int pcm16;
    int stereo;
    audio_prinfo_t *prinfo = &softstate->audio_info.record;

    DPRINTF4(AUD1371M_INTR, "configure_record: %dhz, %d-bit, %s, %d-channel",
	    prinfo->sample_rate,
	    prinfo->precision,
	    (prinfo->encoding == AUDIO_ENCODING_ULAW
	     ? "uLaw"
	     : (prinfo->encoding == AUDIO_ENCODING_LINEAR
		? "linear"
		: "???")),
	    prinfo->channels);

    /* setup ADC recording format */
    pcm16 = (prinfo->precision == AUDIO_PRECISION_16 ||
	     (prinfo->encoding != AUDIO_ENCODING_LINEAR &&
	      prinfo->encoding != AUDIO_ENCODING_LINEAR8));
    stereo = (prinfo->channels == AUDIO_CHANNELS_STEREO);

    serialval = SBPCI_GET32(SBPCI_SERIAL)
	& ~(SBPCI_SERIAL_ADCW16    |
	    SBPCI_SERIAL_ADCSTEREO |
	    SBPCI_SERIAL_NOLOOPADC);
    if (stereo)
	serialval |= SBPCI_SERIAL_ADCSTEREO;
    if (pcm16)
	serialval |= SBPCI_SERIAL_ADCW16;

    SBPCI_PUT32(SBPCI_SERIAL, serialval);

    ENS_SETADCFREQ(softstate, prinfo->sample_rate);
}


/*
 * Setup the ADC sample count register to arrange for a device
 * interrupt after a certain amount of samples have been recorded.
 */
static void
audio1371_set_adc_samplecount(struct audio1371_softstate *softstate,
			      sample_buffer_t *buffer)
{
    int byte_count;
    int sample_count;
    uint32_t serialval;

    byte_count = SAMPLEBUF_SIZE - buffer->used;
    if (byte_count > SAMPLEBUF_INTR_CHUNK)
	byte_count = SAMPLEBUF_INTR_CHUNK;
    sample_count = audio1371_bytes_to_samples(byte_count,
					      &softstate->audio_info.record);

    SBPCI_PUT16(SBPCI_ADC_COUNT, sample_count - 1);

    serialval = SBPCI_GET32(SBPCI_SERIAL);
    SBPCI_PUT32(SBPCI_SERIAL, serialval | SBPCI_SERIAL_ENABLEADC);

    DPRINTF4(AUD1371M_INTR, "set adc samples: %x", sample_count - 1);
}


/*
 * Start recording of samples into an empty sample buffer
 */
static void
audio1371_record_start(struct audio1371_softstate *softstate)
{
    uint32_t controlval;

    mutex_enter(&softstate->intr_mutex);

    /* disable ADC recording channel */
    controlval = SBPCI_GET32(SBPCI_CONTROL);
    SBPCI_PUT32(SBPCI_CONTROL, controlval & ~SBPCI_CTRL_ADC);

    audio1371_configure_recording(softstate);

    /* enable the sample rate converter for the ADC's recording channel */
    ENS_ENABLE_SAMPLE_RATE_CONV(softstate, 3, TRUE, NULL);

    if (audio1371_setup_adc_dma(softstate, &softstate->record_buf)
		== AUDIO1371_SUCCESS) {

	audio1371_set_adc_samplecount(softstate, &softstate->record_buf);

	/* enable the ADC recording channel */
	SBPCI_PUT32(SBPCI_CONTROL, controlval | SBPCI_CTRL_ADC);
    }

    mutex_exit(&softstate->intr_mutex);    
}


/*
 * Stop recording of samples
 */
static void
audio1371_record_stop_unlocked(struct audio1371_softstate *softstate)
{
    uint32_t control;
    sample_buffer_t *buffer;

    /* stop the ADC */
    control = SBPCI_GET32(SBPCI_CONTROL);
    SBPCI_PUT32(SBPCI_CONTROL, control & ~SBPCI_CTRL_ADC);

    /* release DMA resources and return current buffer to free list */
    buffer = &softstate->record_buf;
    if (buffer->dma_bound) {
	ddi_dma_unbind_handle(buffer->dma_handle);
	buffer->dma_bound = FALSE;
    }
}


static void
audio1371_record_stop(struct audio1371_softstate *softstate)
{
    mutex_enter(&softstate->intr_mutex);    
    audio1371_record_stop_unlocked(softstate); 
    mutex_exit(&softstate->intr_mutex);    
}


static int
audio1371_fill_play_buf_unlocked(struct audio1371_softstate *softstate)
{
    sample_buffer_t *buf = &softstate->play_buf;
    int bytes;
    int samples;
    int samples_read;
    int new_data = 0;

    while (buf->used < SAMPLEBUF_SIZE) {
	if (buf->offs + buf->used >= SAMPLEBUF_SIZE) {
	    bytes = SAMPLEBUF_SIZE - buf->used;
	} else {
	    bytes = SAMPLEBUF_SIZE - (buf->offs + buf->used);
	}
	samples = audio1371_bytes_to_samples(bytes, &softstate->audio_info.play);
	DPRINTF4(AUD1371M_INTR, "space for %d samples / %d bytes\n", samples, bytes);

	samples_read = audio_hlp_get_audio(softstate->audio_handle,
		buf->data + (buf->offs+buf->used) % SAMPLEBUF_SIZE,
		0, samples);
	if (samples_read <= 0) {
	    DPRINTF4(AUD1371M_INTR, "got no samples\n");
	    break;
	}
	bytes = audio1371_samples_to_bytes(samples_read, 
					   &softstate->audio_info.play);
#if sparc
	if (softstate->audio_info.play.precision == AUDIO_PRECISION_16)
	    swab(buf->data + (buf->offs+buf->used) % SAMPLEBUF_SIZE,
		 buf->data + (buf->offs+buf->used) % SAMPLEBUF_SIZE,
		 bytes);
#endif

	buf->used += bytes;
	new_data = 1;

	DPRINTF4(AUD1371M_INTR, "got %d samples / %d bytes\n", samples_read, bytes);
    }

    if (new_data) {
	ddi_dma_sync(buf->dma_handle, 0, 0, DDI_DMA_SYNC_FORDEV);
	buf->state = buf->used >= SAMPLEBUF_SIZE
	    ? SAMPLE_BUF_STREAMING_FULL
	    : SAMPLE_BUF_STREAMING_AVAIL;
    }

    return new_data;
}


static void
audio1371_send_record_buf_unlocked(struct audio1371_softstate *softstate)
{
    sample_buffer_t *buf = &softstate->record_buf;
    int bytes;
    int samples;

    ddi_dma_sync(buf->dma_handle, 0, 0, DDI_DMA_SYNC_FORCPU);

    if (buf->offs + buf->used > SAMPLEBUF_SIZE)
	bytes = SAMPLEBUF_SIZE - buf->offs;
    else
	bytes = buf->used;
    samples = audio1371_bytes_to_samples(bytes, &softstate->audio_info.record);
    DPRINTF4(AUD1371M_INTR,
	     "sending %d/%d bytes/samples audio\n", bytes, samples);
#if sparc
    if (softstate->audio_info.record.precision == AUDIO_PRECISION_16)
	swab(buf->data + buf->offs, buf->data + buf->offs, bytes);
#endif
    audio_hlp_send_audio(softstate->audio_handle,
			 buf->data + buf->offs, 0, samples);
    buf->used -= bytes;
    if ((buf->offs += bytes) >= SAMPLEBUF_SIZE)
	buf->offs = 0;
}



#if 0
/*
 * clear pending power level interrupt.
 *
 * to clear the interrupt: read pci power lever 
 * (from pci configuration space at 0xe0) and
 * program the audiopci chip (SBPCI_CONTROL, PDLEV)
 * to the same power level.
 */
static void
audio1371_change_power(struct audio1371_softstate *softstate)
{
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;    
    uint16_t pci_status;
    uint16_t pm_level;
    uint32_t control;

    DPRINTF2(AUD1371M_INTR, "change power");

    if (!conf_handle)
	return;

    pci_status   = pci_config_get16(conf_handle, PCI_CONF_STAT);
    if (pci_status & PCI_STAT_CAP) {
	uint8_t cap_id;
	uint8_t cap_next_ptr;
	uint8_t cap_ptr = pci_config_get8(conf_handle, PCI_CONF_CAP_PTR);

	while (cap_ptr != PCI_CAP_NEXT_PTR_NULL) {
	    cap_id =       pci_config_get8(conf_handle, cap_ptr);
	    cap_next_ptr = pci_config_get8(conf_handle, cap_ptr+1);
	    if (cap_id == PCI_CAP_ID_PM) {
		pm_level = pci_config_get16(conf_handle, cap_ptr + PCI_PMCSR)
		    & PCI_PMCSR_STATE_MASK;

		control = SBPCI_GET32(SBPCI_CONTROL);
		control &= ~SBPCI_CTRL_PDLEV_MASK;
		control |= pm_level << SBPCI_CTRL_PDLEV_SHIFT;
		SBPCI_PUT32(SBPCI_CONTROL, control);
		break;
	    }
	    cap_ptr = cap_next_ptr;
	}
    }
}
#endif



#if defined(sparc)
#undef	EXT_INTR_LOGGING	/* disabled, due to problems (PCI parity  */
				/* errors) on sparc with ens1370 HW, when */
				/* accessing the sample counter register  */
#else
#define	EXT_INTR_LOGGING 1
#endif


/*
 * interrupt handler for the audio1371 device.
 */
static uint_t
audio1371_intr(caddr_t intr_arg)
{
    struct audio1371_softstate *softstate = (struct audio1371_softstate *)intr_arg;
    uint32_t state, control, serial;
    int dacnum;
    uint32_t dma_count;
#if	EXT_INTR_LOGGING
    uint32_t dma_addr;
    uint32_t sample_count;
#endif
    uint32_t new_offs;
    uint32_t old_offs;
    uint32_t bytes_played;
    uint32_t bytes_recorded;
    sample_buffer_t *buffer;

    state  = SBPCI_GET32(SBPCI_STATUS);
    control = SBPCI_GET32(SBPCI_CONTROL);
    serial = SBPCI_GET32(SBPCI_SERIAL);

    DPRINTF4(AUD1371M_INTR,
	     "intr status=%x control=%x serial=%x samplerate=%x time=%lld",
	    state, control, serial,
	    SBPCI_GET32(SBPCI_1371_SAMPLERATE),
	    gethrtime() );

    softstate->intr_count++;

#if	DIAGNOSTICS
    check_pci_status(softstate, "audio1371_intr");
#endif

    if (!(state & SBPCI_STAT_INTR)) {
	/* not an interrupt from the sbpci card */
	return DDI_INTR_UNCLAIMED;
    }

    if (state & (SBPCI_STAT_DAC1 | SBPCI_STAT_DAC2 | SBPCI_STAT_ADC)) {
	/* acknowledge and clear the DAC/ADC interrupt condition */
	if (state & SBPCI_STAT_DAC1) serial &= ~SBPCI_SERIAL_ENABLEDAC1;
	if (state & SBPCI_STAT_DAC2) serial &= ~SBPCI_SERIAL_ENABLEDAC2;
	if (state & SBPCI_STAT_ADC)  serial &= ~SBPCI_SERIAL_ENABLEADC;
	SBPCI_PUT32(SBPCI_SERIAL, serial);

	/*
	 * interrupts are enabled later,
	 * after the sample count register is reprogrammed 
	 */
    }
    

#if	__not_used__
    if (softstate->ens_chip_type != ENS_CHIP_1370 &&
	state & SBPCI_1371_STAT_PWR) {
	/* note: not available on ens1370! */

	/*
	 * power level interrupt pending.
	 *
	 * to clear the interrupt: read pci power lever 
	 * (from pci configuration space at 0xe0) and
	 * program the audiopci chip (SBPCI_CONTROL, PDLEV)
	 * to the same power level
	 */
	audio1371_change_power(softstate);
    }
    if (state & SBPCI_STAT_CCB) {
    }
    if (state & SBPCI_STAT_UART) {
    }
#endif

    if (state & (SBPCI_STAT_DAC1|SBPCI_STAT_DAC2)) {

	select_mempage(softstate, SBPCI_DAC_MEMPG);

	if (state & SBPCI_STAT_DAC1) {
	    dacnum = 1;
	    dma_count = SBPCI_GET32(SBPCI_DAC1_FRAME2);

#if	EXT_INTR_LOGGING
	    dma_addr  = SBPCI_GET32(SBPCI_DAC1_FRAME1);
	    sample_count = SBPCI_GET32(SBPCI_DAC1_COUNT);
#endif
	} else {
	    dacnum = 2;
	    dma_count = SBPCI_GET32(SBPCI_DAC2_FRAME2);

#if	EXT_INTR_LOGGING
	    dma_addr  = SBPCI_GET32(SBPCI_DAC2_FRAME1);
	    sample_count = SBPCI_GET32(SBPCI_DAC2_COUNT);
#endif
	}

#if	EXT_INTR_LOGGING
	DPRINTF4(AUD1371M_INTR,
		 "DAC Frame1 paddr=%x, Frame2 counter=%x, Sample counter=%x",
		 dma_addr, dma_count, sample_count);
#else
	DPRINTF4(AUD1371M_INTR, "DAC Frame2 counter=%x", dma_count);
#endif

#if	0
	/* Dump current sample rate converter RAM */
	{
	    int pg = (state & SBPCI_STAT_DAC1) ? 0 : 4;
	    int pg_end;
	    for (pg_end = pg + 4; pg < pg_end; pg++) {
		select_mempage(softstate, pg);

		DPRINTF4(AUD1371M_INTR,
			 "%x: %08x %08x %08x %08x",
			 pg,
			 SBPCI_GET32(SBPCI_MEMORY),
			 SBPCI_GET32(SBPCI_MEMORY+4),
			 SBPCI_GET32(SBPCI_MEMORY+8),
			 SBPCI_GET32(SBPCI_MEMORY+12));
	    }
	}
#endif

	buffer = &softstate->play_buf;
	mutex_enter(&softstate->intr_mutex);


 	new_offs = (dma_count >> 16) << 2;	/* device DMA read position */
	old_offs = buffer->offs;		/* last read position */

	/* handle wrap around at end of buffer */
	if (new_offs >= old_offs) {
	    bytes_played = new_offs - old_offs;
	} else {
	    bytes_played = new_offs + SAMPLEBUF_SIZE - old_offs;
	}

	DPRINTF4(AUD1371M_INTR,
		 "sbpci_intr: bytes played %x, total avail now %x",
		 bytes_played, buffer->used);

	if (bytes_played >= buffer->used) {
	    DPRINTF4(AUD1371M_INTR, "sbpci_intr: data underflow");

	    /* everything played, we're running out of data: underflow */

	    /* there's nothing left to play, disable dac */
	    audio1371_play_stop_unlocked(softstate, dacnum);

	    mutex_exit(&softstate->intr_mutex);

	    if (softstate->play_timeout_id) {
		untimeout(softstate->play_timeout_id);
		softstate->play_timeout_id = 0;
	    }

	    /* and tell audio helper about end of playback */
	    audio_hlp_play_shutdown(softstate->audio_handle, 0);

	} else {

	    buffer->state = SAMPLE_BUF_STREAMING_AVAIL;
	    buffer->offs = (buffer->offs + bytes_played) % SAMPLEBUF_SIZE;
	    buffer->used -= bytes_played;

	    DPRINTF4(AUD1371M_INTR,
		     "sbpci_intr: offs %x, used %x, state %x",
		     buffer->offs, buffer->used, buffer->state);

	    audio1371_fill_play_buf_unlocked(softstate);

	    audio1371_set_dac_samplecount(softstate, dacnum, buffer);

	    mutex_exit(&softstate->intr_mutex);
	}
    }

    if (state & SBPCI_STAT_ADC) {
	select_mempage(softstate, SBPCI_ADC_MEMPG);
	dma_count = SBPCI_GET32(SBPCI_ADC_FRAME2);

#if	EXT_INTR_LOGGING
	dma_addr  = SBPCI_GET32(SBPCI_ADC_FRAME1);
	sample_count = SBPCI_GET32(SBPCI_ADC_COUNT);

	DPRINTF4(AUD1371M_INTR,
		 "ADC Frame1 paddr=%x, Frame2 counter=%x, Sample counter=%x",
		 dma_addr, dma_count, sample_count);
#else
	DPRINTF4(AUD1371M_INTR, "ADC Frame2 counter=%x", dma_count);
#endif
	
	buffer = &softstate->record_buf;

	mutex_enter(&softstate->intr_mutex);

	new_offs = (dma_count >> 16) << 2;	/* current DMA record offset */
	old_offs = buffer->offs + buffer->used;	/* free space offset in ring */
						/* buffer where recording    */
						/* started */

	if (new_offs >= old_offs) {
	    bytes_recorded = new_offs - old_offs;
	} else {
	    bytes_recorded = new_offs + SAMPLEBUF_SIZE - old_offs;
	}

	buffer->used += bytes_recorded;

	DPRINTF4(AUD1371M_INTR,
		 "sbpci_intr: bytes recorded %x, total avail now %x",
		 bytes_recorded, buffer->used);

	audio1371_set_adc_samplecount(softstate, buffer);

	audio1371_send_record_buf_unlocked(softstate);

	mutex_exit(&softstate->intr_mutex);
    }

    return DDI_INTR_CLAIMED;
}


static int
audio1371_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audio1371_softstate *softstate;
    int status;
    char *getdev_name;

    DPRINTF4(AUD1371M_ATTACH, "attach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_ATTACH:
	break;
    case DDI_RESUME:
    case DDI_PM_RESUME:
    default:
	return DDI_FAILURE;
    }
	
    /*
     * this card is a bus master, reject any slave-only slot
     */
    if (ddi_slaveonly(dip) == DDI_SUCCESS) {
	cmn_err(CE_NOTE, "slot does not support PCI bus-master");
	return DDI_FAILURE;
    }

    /*
     * Make sure we're not using a high level interrupt handler.
     */
    if (ddi_intr_hilevel(dip, 0)) {
	cmn_err(CE_NOTE, "driver does not support high level interrupt");
	cmn_err(CE_NOTE, "You should probably add 'interrupt-priorities=9;' to audio1371.conf");
	return DDI_FAILURE;
    }


    if (ddi_soft_state_zalloc(audio1371_state, instance) != DDI_SUCCESS)
	return DDI_FAILURE;

    softstate = ddi_get_soft_state(audio1371_state, instance);
    softstate->dip = dip;

    if (pci_config_setup(softstate->dip, &softstate->pci_conf_handle) != DDI_SUCCESS) {
	DPRINTF2(AUD1371M_ATTACH, "audio1371%d: pci_config_setup failed", instance);
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (audio1371_pci_setup(softstate) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_ATTACH, "audio1371%d: audio1371_pci_setup failed", instance);
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }


    bcopy(&audio1371_defaults, &softstate->audio_info, sizeof(audio_info_t));
    bcopy(&audio1371_dev_info, &softstate->dev_info, sizeof(audio_device_t));

    /* allow spoofing of audio_device.name in AUDIO_GETDEV ioctl */
    if (ddi_prop_lookup_string(DDI_DEV_T_ANY, dip,
			       DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
			       "ioctl_audio_getdev_name",
			       &getdev_name) == DDI_PROP_SUCCESS) {

	strncpy(softstate->dev_info.name, getdev_name,
		sizeof(softstate->dev_info.name));
	ddi_prop_free(getdev_name);
    }

    sprintf(softstate->dev_info.version, "%02X,%s",
	    softstate->rev, softstate->revstr);



    softstate->ad_info.ad_int_vers = AD_VERS1;
    softstate->ad_info.ad_defaults = &audio1371_defaults;
    bcopy(&audio1371_ch_cap, &softstate->ad_info.ad_play,
	  sizeof(ah_ad_ch_cap_t));
    bcopy(&audio1371_ch_cap, &softstate->ad_info.ad_record,
	  sizeof(ah_ad_ch_cap_t));

    softstate->ad_info.ad_play_comb = audio1371_cap_comb;
    softstate->ad_info.ad_rec_comb = audio1371_cap_comb;
    softstate->ad_info.ad_entry    = &audio1371_entry;
    softstate->ad_info.ad_dev_info = &softstate->dev_info;
    softstate->ad_info.ad_translate_flags =
				AD_MISC_8_P_TRANSLATE | AD_MISC_8_R_TRANSLATE;

    softstate->audio_handle = audio_hlp_attach(dip, cmd, &softstate->ad_info);
    if (!softstate->audio_handle) {
	cmn_err(CE_WARN, "audio1371%d: Failed to attach audio helper module", instance);
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }

    audio_hlp_set_private(softstate->audio_handle, softstate);


    /* init regs handle, before we install the interrupt handler */
    status = ddi_regs_map_setup(dip, 1,
				(caddr_t *)&softstate->regs,
				0, SBPCI_REG_SIZE,
				&dev_attr, &softstate->regs_handle);

    if (status != DDI_SUCCESS) {
	if (status == DDI_REGS_ACC_CONFLICT) {
	    cmn_err(CE_WARN, "regs mapping conflicted with other mapping");
	} else {
	    cmn_err(CE_WARN, "regs mapping failed");
	}
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }

    mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);

    if (ddi_get_iblock_cookie(dip, 0, &softstate->intr_cookie) != DDI_SUCCESS) {
	cmn_err(CE_WARN, "Could not get iblock cookie");
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }
    mutex_init(&softstate->intr_mutex, NULL,
	       MUTEX_DRIVER, (void*)softstate->intr_cookie);

    if (ddi_add_intr(dip, 0,
		     &softstate->intr_cookie, NULL,
		     audio1371_intr, (caddr_t)softstate) != DDI_SUCCESS) {

	cmn_err(CE_WARN, "Could not add interrupt handler");
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }

    /* Allocate buffers for audio playback and recording */

    if (audio1371_alloc_buffer(dip, &softstate->play_buf) != AUDIO1371_SUCCESS) {

	cmn_err(CE_WARN, "Could not allocate playback buf");
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (audio1371_alloc_buffer(dip, &softstate->record_buf) != AUDIO1371_SUCCESS) {

	cmn_err(CE_WARN, "Could not allocate record buf");
	audio1371_cleanup(softstate);
	return DDI_FAILURE;
    }


    softstate->codec_handle = ENS_CODEC_INIT(softstate);

    if (audio1371_reset_card(softstate, &audio1371_defaults, TRUE) != AUDIO1371_SUCCESS) {
	cmn_err(CE_WARN, "cannot reset ensonic 137x sound card");
	audio1371_cleanup(softstate);
	return DDI_FAILURE;	
    }

    DPRINTF4(AUD1371M_ATTACH, "attach\n");
    return DDI_SUCCESS;
}


static int
audio1371_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audio1371_softstate *state = ddi_get_soft_state(audio1371_state, instance);
    int err = DDI_SUCCESS;

    DPRINTF4(AUD1371M_ATTACH, "detach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_DETACH:
	err = audio1371_cleanup(state);
	break;

    default:
	err = DDI_FAILURE;
	break;
    }

    DPRINTF4(AUD1371M_ATTACH, "detach, err=%d\n", err);

    return err;
}


/*
 * Audio Driver functions (for ah_ad_entry_t)
 */

static int    
audio1371_setup(audiohdl_t handle, int stream, int flags)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);

#if 0
    DPRINTF4(AUD1371M_DRIVER, "audio1371_setup, stream %d, flags %x\n", stream, flags);
#else
    audio_hlp_log(handle, 4,
		  "audio1371_setup, stream %d, flags %x\n", stream, flags);
#endif

    mutex_enter(&softstate->intr_mutex);
    audio1371_reset_card(softstate, &softstate->audio_info, FALSE);
    mutex_exit(&softstate->intr_mutex);

    return AUDIO_SUCCESS;
}


static void
audio1371_teardown(audiohdl_t handle, int stream, int dir)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);
    sample_buffer_t *buffer;

#if 0
    DPRINTF4(AUD1371M_DRIVER, "audio1371_teardown, stream %d, dir %d\n", stream, dir);
#else
    audio_hlp_log(handle, 4,
		  "audio1371_teardown, stream %d, dir %d\n", stream, dir);
#endif

    mutex_enter(&softstate->intr_mutex);

    audio1371_stop_device(softstate);

    buffer = &softstate->play_buf;
    if (buffer->dma_bound) {
	ddi_dma_unbind_handle(buffer->dma_handle);
	buffer->dma_bound = FALSE;
    }

    buffer = &softstate->record_buf;
    if (buffer->dma_bound) {
	ddi_dma_unbind_handle(buffer->dma_handle);
	buffer->dma_bound = FALSE;
    }    

    mutex_exit(&softstate->intr_mutex);
}


static int
audio1371_set_config(audiohdl_t handle, int stream,
		     int command, int dir, int arg1, int arg2)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUD1371M_DRIVER, "audio1371_set_config, stream %d, "
	    "command %d, dir %d, arg1 %d, arg2 %d\n",
	    stream, command, dir, arg1, arg2);

    CODEC_SET_CONFIG(softstate->codec_handle, command, dir, arg1, arg2);
    return AUDIO_SUCCESS;
}


static int
audio1371_set_format(audiohdl_t handle, int stream,
		 int dir, int sample_rate, int channels,
		 int precision, int encoding)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);
    audio_prinfo_t *prinfo;

    DPRINTF3(AUD1371M_DRIVER, "audio1371_set_format, stream %d, dir %d, "
	    "sample_rate %d, channels %d, precision %d, encoding %d\n",
	    stream, dir, sample_rate, channels, precision, encoding);

    prinfo = dir == AUDIO_PLAY 
	? &softstate->audio_info.play 
	: &softstate->audio_info.record;

    prinfo->sample_rate = sample_rate;
    prinfo->channels = channels;
    prinfo->precision = precision;
    prinfo->encoding = encoding;
    
    if (dir == AUDIO_PLAY) {
	audio1371_configure_playback(softstate, DEFAULT_PLAYBACK_DAC);
    } else {
	audio1371_configure_recording(softstate);
    }

    return AUDIO_SUCCESS;
}


static void
audio1371_timed_fill_play_buf(void *arg)
{
    struct audio1371_softstate *softstate = arg;
    int state, new_data;

    DPRINTF4(AUD1371M_INTR, "audio1371_timed_fill_play_buf %p\n", softstate);

    mutex_enter(&softstate->intr_mutex);
    softstate->play_timeout_id = 0;

    state = softstate->play_buf.state;
    new_data = audio1371_fill_play_buf_unlocked(softstate);
    if (state == SAMPLE_BUF_IDLE && new_data) {
	/*
	 * we received the first samples when the HW is not currently playing.
	 * Start the HW playbac
	 */
	audio1371_play_start_unlocked(softstate, DEFAULT_PLAYBACK_DAC);
    }

    if (softstate->play_buf.state != SAMPLE_BUF_STREAMING_FULL) {
	softstate->play_timeout_id = 
	    timeout(audio1371_timed_fill_play_buf, softstate, drv_usectohz(20000));
    }
    mutex_exit(&softstate->intr_mutex);
}


static int
audio1371_start_play(audiohdl_t handle, int stream)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);
    int state, new_data;

    DPRINTF4(AUD1371M_DRIVER, "audio1371_start_play, stream %d\n", stream);

    mutex_enter(&softstate->intr_mutex);
    state = softstate->play_buf.state;
    new_data = audio1371_fill_play_buf_unlocked(softstate);

    if (state == SAMPLE_BUF_IDLE && new_data) {
	/*
	 * before calling *_start_play, we were not playing audio data.
	 * the fill_play_buf function has prepared some samples to play,
	 * so start HW playback.
	 */
	audio1371_play_start_unlocked(softstate, DEFAULT_PLAYBACK_DAC);
    }

    if (softstate->play_buf.state != SAMPLE_BUF_STREAMING_FULL
	&& !softstate->play_timeout_id) {
	/*
	 * fill_play_buf didn't find enough data to completely fill the
	 * DMA buffer.  Schedule a timeout function to retry the DMA
	 * fill operation before the HW runs out of sample data
	 */
	softstate->play_timeout_id = 
	    timeout(audio1371_timed_fill_play_buf, softstate, drv_usectohz(20000));
    }
    mutex_exit(&softstate->intr_mutex);

    audio1371_play_pause(softstate, DEFAULT_PLAYBACK_DAC, FALSE);

    return AUDIO_SUCCESS;
}


static void
audio1371_pause_play(audiohdl_t handle, int stream)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUD1371M_DRIVER, "audio1371_pause_play, stream %d\n", stream);

    audio1371_play_pause(softstate, DEFAULT_PLAYBACK_DAC, TRUE);
}


static void
audio1371_stop_play(audiohdl_t handle, int stream)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUD1371M_DRIVER, "audio1371_stop_play, stream %d\n", stream);

    audio1371_play_stop(softstate, DEFAULT_PLAYBACK_DAC);
}


static int
audio1371_get_delay(audiohdl_t handle, int stream)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);
    uint32_t intr_stat;
    uint32_t sample_count;
    uint16_t curr_samp_ct, samp_ct;
    int samples;

    DPRINTF4(AUD1371M_DRIVER, "audio1371_get_delay, stream %d\n", stream);

    mutex_enter(&softstate->intr_mutex);

    if (softstate->play_buf.state >= SAMPLE_BUF_STREAMING_AVAIL) {
	if (softstate->ens_chip_type == ENS_CHIP_1370) {
	    /*
	     * there seems to be a hardware problem with the ensoniq
	     * 1370 chipset, which is triggered when the current sample
	     * counter for the active DAC channel is read; *sometimes*
	     * this seems to generate parity errors on the PCI bus and
	     * that'll panic an UltraSPARC machine.
	     *
	     * The Java VM on Solaris SPARC often triggers the panic
	     * (e.g. the jmplay application).
	     *
	     * For now, avoid accessing the sample counter register
	     * for the active channel (this means we can't report the
	     * current sample position inside the DMA buffer).
	     */
	    curr_samp_ct = samp_ct = 0;

#undef	TEST_ENS1370
#if	TEST_ENS1370
	    /* 
	     * As an experiment, try to access the sample counter of
	     * the *inactive* DAC channel.  Apparently, such an access
	     * does not trigger the PCI parity error and doesn't panic 
	     * an UltraSPARC system.
	     */
	    sample_count = SBPCI_GET32(DEFAULT_PLAYBACK_DAC != 1
				       ? SBPCI_DAC1_COUNT
				       : SBPCI_DAC2_COUNT);
#endif
	} else {
	    sample_count = SBPCI_GET32(DEFAULT_PLAYBACK_DAC == 1
				       ? SBPCI_DAC1_COUNT
				       : SBPCI_DAC2_COUNT);
	    intr_stat =  SBPCI_GET32(SBPCI_STATUS);
	    curr_samp_ct = (sample_count >> 16) & 0xffff;
	    samp_ct = sample_count & 0xffff;

	    /* handle sample counter wraparound on interrupt */
	    if (intr_stat & (DEFAULT_PLAYBACK_DAC == 1
			     ? SBPCI_STAT_DAC1 : SBPCI_STAT_DAC2)
		&& curr_samp_ct >= samp_ct-1)
	    {
		DPRINTF3(AUD1371M_DRIVER,
			 "sample counter wrap around? cur=%x cnt=%x stat=%x\n",
			 curr_samp_ct, samp_ct, intr_stat);
		curr_samp_ct = 0;
	    }
	}
	samples = audio1371_bytes_to_samples(softstate->play_buf.used, 
					     &softstate->audio_info.play);
	if (curr_samp_ct < samp_ct)
	    samples -= (samp_ct - curr_samp_ct);

	/*cmn_err(CE_NOTE, "^%08x, %x, %d\n", sample_count, softstate->play_buf.used, samples);*/
    } else {
	samples = 0;
	/*cmn_err(CE_NOTE, "^stopped, 0\n");*/
    }

    mutex_exit(&softstate->intr_mutex);

    return samples;
}


static int
audio1371_start_record(audiohdl_t handle, int stream)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUD1371M_DRIVER, "audio1371_start_record, stream %d\n", stream);

    audio1371_record_start(softstate);

    return AUDIO_SUCCESS;
}


static void
audio1371_stop_record(audiohdl_t handle, int stream)
{
    struct audio1371_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUD1371M_DRIVER, "audio1371_stop_record, stream %d\n", stream);

    audio1371_record_stop(softstate);
}


static int
audio1371_ioctl(audiohdl_t handle, int channel,
		queue_t *q, mblk_t *mp, int *error)
{
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;
    mblk_t *mp2;
    caddr_t arg = 0;

    if (iocp->ioc_count == TRANSPARENT && (mp2 = mp->b_cont) != NULL)
	/* get the "arg" value from the streams msg */
	bcopy(mp2->b_rptr, &arg, sizeof(arg));

    DPRINTF4(AUD1371M_DRIVER, "audio1371_ioctl ioc_cmd=%x, arg=%p\n", iocp->ioc_cmd, arg);

    return AD_NACK;
}


static int
audio1371_iocdata(audiohdl_t handle, int channel,
		  queue_t *q, mblk_t *mp, int *error)
{
    struct copyresp *copyresp = (struct copyresp *)mp->b_rptr;
    mblk_t *data_mp;
    caddr_t arg = (caddr_t) copyresp->cp_private;

    if (mp->b_cont == NULL) return AD_NACK;
    data_mp = mp->b_cont;

    DPRINTF4(AUD1371M_DRIVER, "audio1371_iocdata cp_cmd=%x, arg=%p\n", copyresp->cp_cmd, arg);

    switch (copyresp->cp_cmd) {
    default:
	return AD_NACK;
    }
}


static audio_info_t audio1371_defaults = {
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_SPEAKER,		/* selected I/O port */
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#endif
	0,			/* Reserved for future use */

	SAMPLEBUF_SIZE,		/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_MICROPHONE,		/* selected I/O port */
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#endif
	0,			/* Reserved for future use */

	SAMPLEBUF_SIZE,		/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    AUDIO_MIN_GAIN,		/* monitor_gain */
    0,				/* output_muted */
#ifdef	AUDIOIO_HAS_REFCNT
    0,				/* driver reference count */
#endif
    {0, 0},			/* reserved fields */
#ifdef	AUDIOIO_HAS_FEATURES
    AUDIO_HWFEATURE_DUPLEX | AUDIO_HWFEATURE_IN2OUT | \
    	AUDIO_HWFEATURE_PLAY | AUDIO_HWFEATURE_RECORD,
    AUDIO_SWFEATURE_MIXER,
    0,
#endif
};


/* 0 term. list of sample rates (used as limits) */
static uint_t audio1371_srs[] = {
    4000, 48000, 0
};


static uint_t audio1371_chs[] = {
    AUDIO_CHANNELS_MONO, 
    AUDIO_CHANNELS_STEREO,
    0
};

static ah_ad_ch_cap_t audio1371_ch_cap = {
    { AD_SRS_FLAG_SR_LIMITS, audio1371_srs },
    audio1371_chs,		/* list of channel types */
    0,				/* interrupt rate ??? */
    2,				/* max channels */
    SAMPLEBUF_SIZE,		/* buffer size */
};


static ah_ad_cap_comb_t audio1371_cap_comb[] = {
    { AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR },  /* 16-bit signed PCM */
    { AUDIO_PRECISION_8,  AUDIO_ENCODING_LINEAR8 }, /* 8-bit unsigned PCM */
    { 0 },
};


/*
 * Audio Driver ops vector definition
 */
static ah_ad_entry_t audio1371_entry = {
    audio1371_setup,
    audio1371_teardown,
    audio1371_set_config,
    audio1371_set_format,
    audio1371_start_play,
    audio1371_pause_play,
    audio1371_stop_play,
    audio1371_start_record,
    audio1371_stop_record,
    audio1371_ioctl,
    audio1371_iocdata,
    audio1371_get_delay,
};


static audio_device_t audio1371_dev_info = {
    DEFAULT_DEVICE_NAME,	/* name */
    "1.0",			/* version */
    "SB16PCI"			/* config */
};
