/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#if	SUNOS >= 509
#include <sys/audio/ac97.h>
#else
#include <sys/ac97.h>
#endif
#include <sys/audio.h>
#include "audioio_compat.h"
#include "ac97_codec.h"
#include "audiovia686.h"

#include <sys/ddi.h>


/*
 * hide gcc2 __attribute__ compiler extension from non-gcc compiler
 * and old gcc compiler
 */
#if     !defined(__GNUC__)
#define __attribute__(x)        /**/
#elif   __GNUC__ < 2
#define __attribute__(x)        /**/
#endif


static int audiov686_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int audiov686_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int audiov686_fill_play_buffers(struct audiov686_softstate *softstate, int unpause);
/*
static int audiov686_prepare_record_buffers(struct audiov686_softstate *softstate);
*/

static audio_info_t audiov686_defaults;


static struct module_info audiov686_minfo = {
    0x0686,		/* module ID number */
    "audv686",		/* module name */
    0,			/* min packet size */
    8192,		/* max packet size */
    48000*4*2,		/* high water mark: ~2 secs of audio */
    0x8000		/* low water mark */
};


static struct qinit audiov686_rqueue = {
    audio_hlp_rput,
    audio_hlp_rsvc,
    audio_hlp_open,
    audio_hlp_close,
    NULL,
    &audiov686_minfo
};

static struct qinit audiov686_wqueue = {
    audio_hlp_wput,
    audio_hlp_wsvc,
    NULL,
    NULL,
    NULL,
    &audiov686_minfo
};


static struct streamtab audiov686_str_info = {
    &audiov686_rqueue,
    &audiov686_wqueue,
    NULL,
    NULL,
};

static struct cb_ops audiov686_cb_ops = {
    nulldev,			/* cb_open */
    nulldev,			/* cb_close */
    nodev,			/* cb_strategy */
    nodev,			/* cb_print */
    nodev,			/* cb_dump */
    nodev,			/* cb_read */
    nodev,			/* cb_write */
    nodev,			/* cb_ioctl */
    nodev,			/* cb_devmap */
    nodev,			/* cb_mmap */
    nodev,			/* cb_segmap */
    nochpoll,			/* cb_chpoll */
    ddi_prop_op,		/* cb_prop_op */
    &audiov686_str_info,	/* cb_str */
    D_NEW|D_MTSAFE|D_64BIT,	/* cb_flag */
    CB_REV,			/* cb_rev */
    nodev,			/* cb_aread */
    nodev,			/* cb_awrite */
};

static struct dev_ops  audiov686_drv_ops = {
    DEVO_REV,
    0,

    audio_hlp_getinfo,
    nulldev,			/* devo_identify */
    nulldev,			/* devo_probe */
    audiov686_attach,
    audiov686_detach,
    nodev,			/* devo_reset */

    &audiov686_cb_ops,
    NULL,			/* devo_bus_ops */
    NULL,			/* devo_power */
};

/*
 * Module linkage information for the kernel.
 */
static struct modldrv audiov686_modldrv = {
    &mod_driverops,		/* Type of module.  This one is a driver */
    "via 686 audio driver",
    &audiov686_drv_ops		/* driver ops */
};


static struct modlinkage audiov686_modlinkage = {
    MODREV_1,
    { &audiov686_modldrv, NULL },
};


/*
 * This describes the VIA controller's DMA engines to the OS at hardware level
 * 
 * The first structure is for the "SGD table".
 * These must be in contiguous physical memory.
 *
 * The VIA686 accesses SGD memory in 8 bytes quantities (=> minxfer == 8)
 * and the whole sgd table has a fixed size of 32*8 == 256 bytes. An sgd table
 * must be allocated in contiguous physical memory (=> s/g length == 1).
 * 
 * The typical DMA transfer is a burst of 8 longwords (=> DMA burstsizes
 * has 0x20 bit set; I added 0x10, 0x8 and 0x4, too)
 */
static ddi_dma_attr_t sgd_dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0x0000ffff,			/* DMA counter register */
    8,				/* DMA address alignment */
    				/*   sgd list must be aligned on a 8-byte boundary */
    0x3c,			/* DMA burstsizes ?? */
    8,				/* minxfer effective DMA size */
    0x0000ffff,			/* max DMA xfer size */
    0x00000fff,			/* segment boundary*/
    1,				/* s/g length */
    8,				/* granularity of device */
				/*   size of a sgd entry?? */
    0				/* flags */
};


/*
 * DMA attributes for sample data.  No need for physical contiguous
 * memory.
 *
 * The VIA686 accesses sample data in longword (4 bytes) quantities
 * (=> minxfer == 4) and has a 16-bit word DMA counter (=> max DMA
 * xfer size == 0xffff * 2).  It does support scatter/gatter DMA
 * transfers from upto 32 memory locations (=> s/g length == 32).
 * 
 * The typical DMA transfer is a burst of 8 longwords (=> DMA burstsizes
 * has 0x20 bit set; I added 0x10, 0x8 and 0x4, too)
 */
static ddi_dma_attr_t dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0x00ffffff,			/* DMA counter register */
    2,				/* DMA address alignment */
    0x3c,			/* DMA burstsizes */
    4,				/* minxfer effective DMA size */
    0x00ffffff,			/* max DMA xfer size */
    0x00001fff,			/* segment boundary*/
    V686_NUM_SGD,		/* s/g length */
    4,				/* granularity of device */
				/*   (VIA AC97 dma operates on longwords) */
    0				/* flags */
};


/* This describes things like byteorder internally for the ICH */
static struct ddi_device_acc_attr dev_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_NEVERSWAP_ACC,
    DDI_STRICTORDER_ACC
};


/* available log levels */
#define	AUDV686_LOG_L0	0		/* warnings, console & syslog buffer */
#define	AUDV686_LOG_L1	1		/* errors, syslog buffer */
#define	AUDV686_LOG_L2	2		/* recoverable errors, debug only */
#define	AUDV686_LOG_L3	3		/* interesting data, debug only */
#define	AUDV686_LOG_L4	4		/* tracing, debug only */


/* available log masks */
#define	AUDV686M_AC97	0x00000020	/* AC97 register access */
#define	AUDV686M_CODEC	0x00000010	/* CODEC IO */
#define	AUDV686M_INTR	0x00000008	/* interrupt handling */
#define	AUDV686M_DRIVER	0x00000004	/* audio driver functions */
#define	AUDV686M_ATTACH	0x00000002	/* card setup at attach time */
#define	AUDV686M_MODS	0x00000001	/* _init, _info, _fini */
#define	AUDV686M_ALL	(~0u)

uint_t audiov686_errlevel = AUDV686_LOG_L0;
uint_t audiov686_errmask =  AUDV686M_ALL;


/* logging for log levels 4..0 */
#if	DEBUG
#define	DPRINTF4(...)	audiov686_dprintf4(__VA_ARGS__)
#define	DPRINTF3(...)	audiov686_dprintf3(__VA_ARGS__)
#define	DPRINTF2(...)	audiov686_dprintf2(__VA_ARGS__)
#else
#define	DPRINTF4(...)	/**/
#define	DPRINTF3(...)	/**/
#define	DPRINTF2(...)	/**/
#endif
#define	DPRINTF1(...)	audiov686_dprintf1(__VA_ARGS__)
#define	DPRINTF0(...)	audiov686_dprintf0(__VA_ARGS__)


static void
audiov686_vprintf(uint_t level, uint_t mask, char *fmt, va_list args)
{
    if (level <= audiov686_errlevel && (mask & audiov686_errmask))
	vcmn_err(CE_CONT, fmt, args);
}


#ifdef	DEBUG
void __attribute__ ((format (__printf__, 2, 3)))
audiov686_dprintf4(uint_t mask, char *fmt, ...)

{
    va_list args;
    va_start(args, fmt);
    audiov686_vprintf(AUDV686_LOG_L4, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audiov686_dprintf3(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov686_vprintf(AUDV686_LOG_L3, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audiov686_dprintf2(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov686_vprintf(AUDV686_LOG_L2, mask, fmt, args);
    va_end(args);
}
#endif

void __attribute__ ((format (__printf__, 2, 3)))
audiov686_dprintf1(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov686_vprintf(AUDV686_LOG_L1, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audiov686_dprintf0(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov686_vprintf(AUDV686_LOG_L0, mask, fmt, args);
    va_end(args);
}

#if 0
/* HardCore debug function
 */
static void print_gds(audiov686_softstate_t *softstate)
{
   int i;
   v686_sgd_t * sgd = softstate->sgd;
   v686_buf_descr_t * sgdp = sgd->pcm_out_sgd;
   for(i=0;i<V686_NUM_SGD;i++){
    DPRINTF4(AUDV686M_ALL, "sgd[%d]=(0x%x,%d,%s,%s,%s)\n",
	   i,
	   sgdp[i].bd_base,
	   sgdp[i].bd_len,
	   sgdp[i].bd_stop ? "STOP" : "-",
	   sgdp[i].bd_flag ? "FLAG" : "-",
	   sgdp[i].bd_eol  ? "EOL"  : "-");
   }
}
#endif

/* 
 * Global softstate for multiple via 686 audio controller instances
 */
static void *audiov686_state;


/* 
 * Sun driver module initialization and finalization code 
 */
int 
_init(void)
{
    int err;

    err = ddi_soft_state_init(&audiov686_state,
			      sizeof(struct audiov686_softstate), 1);
    if (err == 0) {
	if ((err = mod_install(&audiov686_modlinkage)) != 0)
	    ddi_soft_state_fini(&audiov686_state);
    }

    DPRINTF4(AUDV686M_MODS, "_init, err=%d\n", err);

    return err;
}


int
_fini(void)
{
    int err;
    
    if ((err = mod_remove(&audiov686_modlinkage)) == 0)
	ddi_soft_state_fini(&audiov686_state);

    DPRINTF4(AUDV686M_MODS, "_fini, err=%d\n", err);

    return err;
}


int
_info(struct modinfo *modinfop)
{
    return mod_info(&audiov686_modlinkage, modinfop);
}

/*
 * Allocate a list of DMA memory buffers (for playback or record)
 * and place the 8K chunks on the buffers free list.
 */
static int
audiov686_alloc_sample_buffer(struct audiov686_softstate *softstate,
			      sample_buffer_t *buf)
{
    int offs;
    struct sgd_chunk *chunk;

    for (offs = 0; offs < SAMPLEBUF_SIZE; offs += SAMPLEBUF_INTR_CHUNK) {

	chunk = kmem_zalloc(sizeof(struct sgd_chunk), KM_NOSLEEP);
	if (chunk == NULL)
	    return AUDIOV686_FAILURE;

	chunk->next = buf->free;
	buf->free = chunk;

	chunk->dma_bound = FALSE;

	if (ddi_dma_alloc_handle(softstate->dip,
				 &dma_attr,
				 DDI_DMA_DONTWAIT, NULL,
				 &chunk->dma_handle) != DDI_SUCCESS)
	    return AUDIOV686_FAILURE;
    
	if (ddi_dma_mem_alloc(chunk->dma_handle,
			      SAMPLEBUF_INTR_CHUNK,
			      &dev_attr,
			      DDI_DMA_STREAMING,
			      DDI_DMA_DONTWAIT, NULL,
			      &chunk->data, &chunk->reallen,
			      &chunk->acc_handle) != DDI_SUCCESS) {
	    DPRINTF1(AUDV686M_ATTACH, "alloc dma sample buffer failed");
	    return AUDIOV686_FAILURE;
	}
    }

    return AUDIOV686_SUCCESS;
}


/*
 * Release the DMA memory sample buffers allocated with
 * audiov686_alloc_sample_buffer()
 */
static int
audiov686_free_sample_buffer(struct audiov686_softstate *softstate,
			     sample_buffer_t *buf)
{
    struct sgd_chunk *chunk, *next;

    for (chunk = buf->free; chunk != NULL; chunk = next) {
	next = chunk->next;

	if (chunk->acc_handle) {
	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }

	    ddi_dma_mem_free(&chunk->acc_handle);
	}

	if (chunk->dma_handle) {
	    ddi_dma_free_handle(&chunk->dma_handle);
	}
	
	kmem_free(chunk, sizeof(struct sgd_chunk));
    }

    return AUDIOV686_SUCCESS;
}


/*
 * Configure the PCI device, so that the device is ready to 
 * accept I/O requests.  This includes enabling PCI busmaster,
 * and anabling access to the device's I/O registers.
 */
static int
audiov686_pci_setup(struct audiov686_softstate *softstate)
{
    dev_info_t *dip = softstate->dip;
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;    
    uint16_t pci_status;
    uint16_t pci_vendorid;
    uint16_t pci_deviceid;
    uint8_t  pci_revision;
    uint8_t  pci_intrinfo;
    uint16_t pci_commandreg;
    char *revstr = "";

    pci_status   = pci_config_get16(conf_handle, PCI_CONF_STAT);
    pci_vendorid = pci_config_get16(conf_handle, PCI_CONF_VENID);
    pci_deviceid = pci_config_get16(conf_handle, PCI_CONF_DEVID);
    pci_revision = pci_config_get8 (conf_handle, PCI_CONF_REVID);
    pci_intrinfo = pci_config_get8 (conf_handle, PCI_CONF_ILINE);    
    
    /* try to find out a nice hardware name for the given pci hw */
    if (pci_vendorid == 0x1106) {
	/* intel: */
	switch (pci_deviceid) {
	case 0x3058: revstr = "VT82C686A"; break;
	}
    }
    softstate->rev = pci_revision;
    softstate->revstr = revstr;

    /* Announce the hw we're attaching to */
    ddi_report_dev(dip);
    cmn_err(CE_CONT, "?PCI-device: %s, audiov686%d, %s%s(rev. 0x%02x) at irq %d\n",
	    ddi_node_name(dip),
	    ddi_get_instance(dip), 
	    revstr, revstr[0] ? " " : "", 
	    pci_revision,
	    pci_intrinfo);

#if i386
    /*
     * reject interrupt line 255, per PCI spec 6.2.4, footnote 43,
     * this means "unknown" or "not connected" to the interrupt
     * controller on x86 based PCs.
     * Sun drivers (iprb) reject iline 0 as well.
     */
    if (pci_intrinfo == 0 || pci_intrinfo == 255) {
	cmn_err(CE_WARN, "interrupt line value out of range: %d", pci_intrinfo);
	return AUDIOV686_FAILURE;
    }
#endif
    
    pci_commandreg = pci_config_get16(conf_handle, PCI_CONF_COMM);
    DPRINTF4(AUDV686M_ATTACH, "pci command register 0x%x %s%s%s%s",
	    pci_commandreg,
	    pci_commandreg & PCI_COMM_SERR_ENABLE ? " serr#_en"  : "",
	    pci_commandreg & PCI_COMM_ME  ? " busmaster" : "",
	    pci_commandreg & PCI_COMM_MAE ? " mem_access" : "",
	    pci_commandreg & PCI_COMM_IO  ? " io_access" : "");
    if ((pci_commandreg & (PCI_COMM_IO|PCI_COMM_ME)) == (PCI_COMM_IO|PCI_COMM_ME)) {
	DPRINTF4(AUDV686M_ATTACH, "busmastering and io_access already enabled");
    } else {
	DPRINTF4(AUDV686M_ATTACH, "Enabling PCI I/O and bus mastering access");
	pci_config_put16(conf_handle, PCI_CONF_COMM, pci_commandreg | PCI_COMM_IO | PCI_COMM_ME);
    }

    return AUDIOV686_SUCCESS;
}


static int
audiov686_wait_for_codec_idle(struct audiov686_softstate *softstate)
{
    int loop;
    uint32_t value = 0;

    for (loop = 100; --loop >= 0; ) {
        value = V686_GET32(V686_AC97_COMMAND);
        if ((value & V686_AC97_CMD_CONTROLLER_BUSY) == 0) {
            /* codec is idle now */
            return AUDIOV686_SUCCESS;
        }
        drv_usecwait(10);
    }

    DPRINTF1(AUDV686M_ALL, "timeout, codec not idle [%x]", value);
    return AUDIOV686_FAILURE;
}

static int
audiov686_wait_for_codecdata(struct audiov686_softstate *softstate, uint16_t *data)
{
    int loop;
    uint32_t value = 0;

    for (loop = 100; --loop >= 0; ) {
        value = V686_GET32(V686_AC97_COMMAND);
        if ((value & (V686_AC97_CMD_PRIMARY_VALID|V686_AC97_CMD_CONTROLLER_BUSY))
	    	== V686_AC97_CMD_PRIMARY_VALID) {
           /* codec not busy any more + our requested data is ready */
            *data = value & V686_AC97_CMD_DATA_MASK;
            return AUDIOV686_SUCCESS;
        }
        drv_usecwait(10);
    }

    DPRINTF1(AUDV686M_ALL, "timeout waiting got codec read [%x]", value);
    return AUDIOV686_FAILURE;
}

/*
 * Read an AC97 primary codec mixer register
 *
 * Returns TRUE on success with the data read returned via the
 * *data pointer.  Returns FALSE in case the primary codec is not
 * ready any more for some reason.
 */
static int
audiov686_ac97_read(void *closure, uint16_t sub_addr, uint16_t *data)
{
    struct audiov686_softstate *softstate = closure;
    uint32_t codecreg = 0;
    uint16_t subreg;

    codecreg = V686_AC97_CMD_PRIMARY_CODEC |
	V686_AC97_CMD_PRIMARY_VALID |
	V686_AC97_CMD_READ_ACCESS |
	V686_AC97_CMD_REG_IDX(sub_addr);
    
    if (audiov686_wait_for_codec_idle(softstate) != AUDIOV686_SUCCESS) {
	return FALSE;
    }

    V686_PUT32(V686_AC97_COMMAND, codecreg);

    /* wait and read the codec register */
    if (audiov686_wait_for_codecdata(softstate, &subreg) != AUDIOV686_SUCCESS) {
        return FALSE;
    }
    *data = subreg;

    return TRUE;
}


/*
 * Write to an AC97 primary codec mixer register.
 *
 * Returns TRUE on success. Returns FALSE in
 * case the primary codec is not ready any more for some reason.
 */
static int
audiov686_ac97_write(void *closure, uint16_t sub_addr, uint16_t value)
{
    struct audiov686_softstate *softstate = closure;
    uint32_t codecvalue;

    codecvalue = V686_AC97_CMD_REG_IDX(sub_addr) |
	(value & V686_AC97_CMD_DATA_MASK);

    codecvalue |= V686_AC97_CMD_PRIMARY_CODEC | V686_AC97_CMD_WRITE_ACCESS;

    if (audiov686_wait_for_codec_idle(softstate) != AUDIOV686_SUCCESS) {
	return FALSE;
    }

    V686_PUT32(V686_AC97_COMMAND, codecvalue);

    return TRUE;
}


/*
 * Completely disable all VIA audio controller DMA
 */
static void
audiov686_stop_dma(struct audiov686_softstate *softstate)
{
    int i;
    V686_PUT8(V686_SGD_RDCHN_CTRL, V686_SGD_CTRL_TERMINATE);
/*    V686_PUT8(V8233_SGD_WRCHN_CTRL, V8233_SGD_CTRL_TERMINATE); */
  for(i=0;i<10000;i++){
    int status;
    drv_usecwait(50);
    status = V686_GET8( V686_SGD_RDCHN_STAT );
    if (  (status & V686_SGD_STAT_ACTIVE) == 0 ){
      break;
    } 
  }
  if ( i == 10000 ){
    DPRINTF0(AUDV686M_ALL, "Read channel can not stop!");
  }

    /* 
     * NOTE:
     * It is necessary to ack possible pending interrupt from sound card
     * otherwise subsequent playback  won't start at all.
     */
    V686_PUT8(V686_SGD_RDCHN_STAT, 
	       V686_SGD_STAT_STOPPED|V686_SGD_STAT_EOL|V686_SGD_STAT_FLAG);
}


/*
 * reset the VIA audiocontroller's AC'97 link and reset the AC'97
 * codec to default values
 */
static int
audiov686_reset_card(struct audiov686_softstate *softstate,
		     audio_info_t		*audio_info,
		     int			 verbose)
{
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;
    uint8_t ac97_ctrl;
    uint8_t ac97_stat = 0;
    int loop;

    ac97_ctrl = pci_config_get8(conf_handle, PCI_CONF_V686_AC97_CTRL);
    if (ac97_ctrl & PCI_V686_ACCTL_ACRST) {
	/* reset is de-asserted, perform a warm AC97 reset:
	 * drive SYNC high for a minimum of 1us */
	pci_config_put8(conf_handle, PCI_CONF_V686_AC97_CTRL, 
			PCI_V686_ACCTL_ENAC97 |
			PCI_V686_ACCTL_ACRST  | 
			PCI_V686_ACCTL_RSYNC_HI);
	drv_usecwait(10);
	pci_config_put8(conf_handle, PCI_CONF_V686_AC97_CTRL,
			PCI_V686_ACCTL_ENAC97 |
			PCI_V686_ACCTL_ACRST);
    } else {
	/* cold AC97 reset: 
	 * assert reset# low for the minimum specified time ... */
	pci_config_put8(conf_handle, PCI_CONF_V686_AC97_CTRL, 0);
	drv_usecwait(100);
	/* ... then de-asserting reset# high */
	pci_config_put8(conf_handle, PCI_CONF_V686_AC97_CTRL,
			PCI_V686_ACCTL_ACRST);
    }

    pci_config_put8(conf_handle, PCI_CONF_V686_AC97_CTRL,
		    PCI_V686_ACCTL_ENAC97 |
		    PCI_V686_ACCTL_ACRST | 
		    PCI_V686_ACCTL_VSR_ON_DEMAND |
		    PCI_V686_ACCTL_SGD_READ_CHANNEL);

    /* wait for ac97 codec ready status */
    for (loop = 6000; --loop >= 0; ) {
	ac97_stat = pci_config_get8(conf_handle, PCI_CONF_V686_AC97_STAT);
	if (ac97_stat & PCI_V686_ACSTAT_PRIMARY_CODEC_READY)
	    break;
	drv_usecwait(100);
    }
    printf("AC97 cold/warm reset, initial ac97 ctl %02x, ac97 status %02x, loop %d\n",
	   ac97_ctrl, ac97_stat, loop);

    if (loop < 0) {
	DPRINTF2(AUDV686M_ATTACH,
		 "AC97 cold/warm reset still in progress? [%02x,%02x]\n",
		 ac97_ctrl, ac97_stat);
	return AUDIOV686_FAILURE;
    }

    /* no DMA (not yet) */
    audiov686_stop_dma(softstate);

    if (!CODEC_RESET(softstate->codec_handle, audio_info, verbose)) {
        DPRINTF2(AUDV686M_CODEC, "Audio CODEC reset failed");
        return AUDIOV686_FAILURE;
    }

    audiov686_ac97_read (softstate, AC97_EXTENDED_AUDIO_REGISTER, &softstate->ac97_ext_features);
#if 0
#warning software SRC forced
    /* force the use of the software sample rate converter */
    softstate->ac97_ext_features = 0;
#endif

    return AUDIOV686_SUCCESS;
}
/*
 * Allocate DMA resources for the VIA's SGD buffer descriptor lists.  We
 * allocate adjacent DMA memory in one physical page for all three DMA
 * engines available with the VIA audio controller.
 */
static int
audiov686_setup_dma(struct audiov686_softstate *softstate)
{
    int i, ret;
    uint_t ccount;
    ddi_dma_cookie_t cookie;

    softstate->sgd_dma_bound = FALSE;

    if ((ret = ddi_dma_alloc_handle(softstate->dip, &sgd_dma_attr,
				    DDI_DMA_DONTWAIT, NULL,
				    &softstate->sgd_dma_handle)) != DDI_SUCCESS) {
	DPRINTF4(AUDV686M_ATTACH, "setup_dma: alloc handle failed (%d)\n", ret);
	return AUDIOV686_FAILURE;
    }

    if (ddi_dma_mem_alloc(softstate->sgd_dma_handle,
			  sizeof(v686_sgd_t),
			  &dev_attr,
			  DDI_DMA_CONSISTENT,
			  DDI_DMA_DONTWAIT, NULL,
			  (caddr_t*) &softstate->sgd,
			  &softstate->sgdlen,
			  &softstate->sgd_acc_handle) != DDI_SUCCESS) {
	DPRINTF4(AUDV686M_ATTACH, "setup_dma: mem alloc failed\n");
	return AUDIOV686_FAILURE;
    }

    if ((ret = ddi_dma_addr_bind_handle(softstate->sgd_dma_handle,
					NULL,
					(caddr_t)softstate->sgd, softstate->sgdlen,
					DDI_DMA_RDWR|DDI_DMA_CONSISTENT,
					DDI_DMA_DONTWAIT, NULL,
					&cookie, &ccount)) != DDI_DMA_MAPPED) {
	DPRINTF4(AUDV686M_ATTACH, "setup_dma: bind failed (%d)\n", ret);
	return AUDIOV686_FAILURE;
    }

    /*
     * We should receive exactly one physical page mapping
     * (=contiguous physical memory)
     */
    if (ccount != 1) {
	DPRINTF1(AUDV686M_ATTACH, "unexpected dma cookie count > 1 (%u)\n", ccount);
	return AUDIOV686_FAILURE;
    }

    softstate->sgd_dma_bound = TRUE;

    DPRINTF4(AUDV686M_ATTACH, "sgd at physical address %p\n",
	     (void*) (long) cookie.dmac_address);

    softstate->sgd_paddr = (v686_sgd_t*) (long) cookie.dmac_address;

    bzero(softstate->sgd, sizeof(v686_sgd_t));
    for (i = 0; i < V686_NUM_SGD; i++) {
	softstate->sgd->pcm_out_sgd[i].bd_flag = TRUE;
	softstate->sgd->pcm_out_sgd[i].bd_stop = TRUE;

	softstate->sgd->pcm_in_sgd[i].bd_flag = TRUE;
	softstate->sgd->pcm_in_sgd[i].bd_stop = TRUE;

    }

    softstate->sgd->pcm_out_sgd[V686_NUM_SGD-1].bd_eol = TRUE;
    softstate->sgd->pcm_in_sgd[V686_NUM_SGD-1].bd_eol = TRUE;
    return AUDIOV686_SUCCESS;
}


/*
 * Cleanup all resources allocated by the driver during attach time
 * and during run-time, in reverse order.
 */
static int
audiov686_cleanup(struct audiov686_softstate *softstate)
{
    if (softstate->audio_handle) {
	if (audio_hlp_detach(softstate->audio_handle, DDI_DETACH) != DDI_SUCCESS)
	    return DDI_FAILURE;
    }

    if (softstate->codec_handle) {
        CODEC_DELETE(softstate->codec_handle);
        softstate->codec_handle = NULL;
    }

    if (softstate->sgd_acc_handle != NULL) {

	/*
	 * just to be sure, turn off DMA, before releaseing the DMA
	 * sample buffers
	 */
	audiov686_stop_dma(softstate);

	/* free audio sample buffer DMA memory */
	audiov686_free_sample_buffer(softstate, &softstate->play_buf);
	audiov686_free_sample_buffer(softstate, &softstate->record_buf);

	/* free SGD DMA memory for the three VIA DMA engines */
	if (softstate->sgd_dma_bound)
	    ddi_dma_unbind_handle(softstate->sgd_dma_handle);
	ddi_dma_mem_free(&softstate->sgd_acc_handle);
	softstate->sgd = NULL;
    }

    if (softstate->sgd_dma_handle) {
	ddi_dma_free_handle(&softstate->sgd_dma_handle);
    }

    /*
     * remove interrupt handler before we free the card's mapped registers
     * otherwise an interrupt could access softstate->regs_handle in our
     * interrupt handler
     */
    if (softstate->intr_cookie != NULL) {
	ddi_remove_intr(softstate->dip, 0, softstate->intr_cookie);
    }

    if (softstate->regs_handle != NULL) {
	ddi_regs_map_free(&softstate->regs_handle);
    }

    if (softstate->pci_conf_handle != NULL) {
	pci_config_teardown(&softstate->pci_conf_handle);
    }

    mutex_destroy(&softstate->instance_mutex);
    mutex_destroy(&softstate->intr_mutex);

    ddi_soft_state_free(audiov686_state, ddi_get_instance(softstate->dip));

    return DDI_SUCCESS;
}

/*
 * Playback of a chunk of sample data has finished, release the DMA buffer
 * to the free list.
 */
static void
audiov686_release_play_buf_head(struct audiov686_softstate *softstate,
				sample_buffer_t *buf)
{
    struct sgd_chunk *chunk;
    v686_buf_descr_t *bd;

    DPRINTF4(AUDV686M_INTR, "release enter\n");
    bd = softstate->sgd->pcm_out_sgd + buf->head;
    bd->bd_len  = 0;
    bd->bd_stop = TRUE;
    bd->bd_flag = TRUE;

    if ((chunk = buf->active[buf->head]) != NULL) {
	DPRINTF4(AUDV686M_INTR, "free chunk at index %d\n", buf->head);
	if (chunk->dma_bound) {
	    ddi_dma_unbind_handle(chunk->dma_handle);
	    chunk->dma_bound = FALSE;
	}
	chunk->next = buf->free;
	buf->free = chunk;
	buf->active[buf->head] = NULL;
    }
    ++buf->buffers_available;
    if (++buf->head >= V686_NUM_SGD)
	buf->head = 0;

    DPRINTF4(AUDV686M_INTR, "release exit\n");
}


#if 0
/*
 * Recoding to a chunk of sample data at the front of the DMA buffer
 * descriptor list has finished.  Send the recorded audio samples to
 * the audio driver framework and release the buffer to the free list.
 */
static void
audiov686_release_record_buf_head(struct audiov686_softstate *softstate,
				  sample_buffer_t *buf)
{
    struct sgd_chunk *chunk;
    int byte_to_sample_shift;

    byte_to_sample_shift = 0;
    if (softstate->audio_info.record.channels == 2) byte_to_sample_shift++;
    if (softstate->audio_info.record.precision == 16) byte_to_sample_shift++;

    if ((chunk = buf->active[buf->head]) != NULL) {
	DPRINTF4(AUDV686M_INTR, "free chunk at index %d, %d samples\n",
		 buf->head,
		 SAMPLEBUF_INTR_CHUNK >> byte_to_sample_shift);
	if (chunk->dma_bound) {
	    ddi_dma_unbind_handle(chunk->dma_handle);
	    chunk->dma_bound = FALSE;
	}

	audio_hlp_send_audio(softstate->audio_handle, chunk->data, 0,
			     SAMPLEBUF_INTR_CHUNK >> byte_to_sample_shift);

	chunk->next = buf->free;
	buf->free = chunk;
	buf->active[buf->head] = NULL;
    }
    ++buf->buffers_available;
    if (++buf->head >= V686_NUM_SGD)
	buf->head = 0;
}
#endif


/*
 * interrupt handler for the audiov686 device.
 * NOTE: It is Hardware Interrupt Handler. It may call only restricted
 * kernel function set. It may not sleep. It should be 
 * short and fast.
 */
static uint_t
audiov686_intr(caddr_t intr_arg)
{
    struct audiov686_softstate *softstate = (void*)intr_arg;

    sample_buffer_t *buf;
    uint8_t play_stat;
    int claimed = DDI_INTR_UNCLAIMED;

    play_stat = V686_GET8(V686_SGD_RDCHN_STAT);
    play_stat &= (V686_SGD_STAT_STOPPED|V686_SGD_STAT_FLAG|V686_SGD_STAT_EOL);
    if (play_stat) {
	claimed = DDI_INTR_CLAIMED;

	mutex_enter(&softstate->intr_mutex);
	buf = &(softstate->play_buf);

        buf->cur_idx = (uint8_t)
	    ((V686_GET32(V686_SGD_RDCHN_PTR)
	      -(long)(softstate->sgd_paddr->pcm_out_sgd))/sizeof(v686_buf_descr_t));

	DPRINTF4(AUDV686M_INTR, "intr stat 0x%x, idx %d, head %d, tail %d\n",
		 play_stat, buf->cur_idx, buf->head, buf->tail); 
	V686_PUT8(V686_SGD_RDCHN_STAT, play_stat);

	/* buffer underrun ( buffer list empty)? */
	if ( buf->cur_idx == buf->tail ){

	   /* if yes - > pause */
	   V686_PUT8(V686_SGD_RDCHN_CTRL, V686_SGD_CTRL_PAUSE);
	   DPRINTF4(AUDV686M_INTR, "buffer underrun or eof"); 
	}

	mutex_exit(&softstate->intr_mutex);
    }

    /* TODO: recording stuff */

#if 0
    DPRINTF4(AUDV686M_INTR, "audiov686_intr release: claimed %s\n",
	    claimed == DDI_INTR_CLAIMED ? "yes" : "no" );
#endif
    return claimed;
}


static audio_device_t audiov686_dev_info;
static ah_ad_entry_t audiov686_entry;
static ah_ad_ch_cap_t audiov686_ch_cap;
static ah_ad_cap_comb_t audiov686_cap_comb[];
static ah_ad_sample_rates_t audiov686_sample_rates_48k;

static int
audiov686_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audiov686_softstate *softstate;
    int status;
    char *getdev_name;

    DPRINTF4(AUDV686M_ATTACH, "attach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_ATTACH:
	break;

    case DDI_RESUME:
    case DDI_PM_RESUME:
    default:
	return DDI_FAILURE;
    }

    /*
     * this card is a bus master, reject any slave-only slot
     */
    if (ddi_slaveonly(dip) == DDI_SUCCESS) {
        cmn_err(CE_NOTE, "slot does not support PCI bus-master");
        return DDI_FAILURE;
    }

    /*
     * Make sure we're not using a high level interrupt handler.
     */
    if (ddi_intr_hilevel(dip, 0)) {
	cmn_err(CE_NOTE, "Driver does not support high level interrupt.");
	cmn_err(CE_NOTE, "You should probably add 'interrupt-priorities=9;' to audiov686.conf");
	return DDI_FAILURE;
    }

    if (ddi_soft_state_zalloc(audiov686_state, instance) != DDI_SUCCESS)
	return DDI_FAILURE;

    softstate = ddi_get_soft_state(audiov686_state, instance);
    softstate->dip = dip;

    if (pci_config_setup(softstate->dip, &softstate->pci_conf_handle) != DDI_SUCCESS) {
	DPRINTF4(AUDV686M_ATTACH, "audiov686%d: pci_config_setup failed", instance);
	audiov686_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (audiov686_pci_setup(softstate) != AUDIOV686_SUCCESS) {
	DPRINTF4(AUDV686M_ATTACH, "audiov686%d: audiov686_pci_setup failed", instance);
	audiov686_cleanup(softstate);
	return DDI_FAILURE;	
    }

    bcopy(&audiov686_defaults, &softstate->audio_info, sizeof(audio_info_t));
    bcopy(&audiov686_dev_info, &softstate->dev_info, sizeof(audio_device_t));

    /* allow spoofing of audio_device.name in AUDIO_GETDEV ioctl */
    if (ddi_prop_lookup_string(DDI_DEV_T_ANY, dip,
			       DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
			       "ioctl_audio_getdev_name",
			       &getdev_name) == DDI_PROP_SUCCESS) {

	strncpy(softstate->dev_info.name, getdev_name,
		sizeof(softstate->dev_info.name));
	ddi_prop_free(getdev_name);
    }

    sprintf(softstate->dev_info.version, "%02X,%s",
	    softstate->rev, softstate->revstr);

    softstate->ac97_mic_20db_boost =
	ddi_getprop(DDI_DEV_T_ANY, dip,
		    DDI_PROP_DONTPASS | DDI_PROP_CANSLEEP | DDI_PROP_NOTPROM,
		    "mic20db", TRUE);

    /*
     * sun's ac97.h header file contains a wrong value for
     * PCMOVR_RIGHT_0dB_GAIN (16 instead of 8), we avoid this define.
    */
    softstate->ac97_pcmvol =
	ddi_getprop(DDI_DEV_T_ANY, dip,
		    DDI_PROP_DONTPASS | DDI_PROP_CANSLEEP | DDI_PROP_NOTPROM,
		    "pcmvol", 8 /*PCMOVR_RIGHT_0dB_GAIN*/) & PCMOVR_RIGHT_GAIN_MASK;

    softstate->ad_info.ad_int_vers = AD_VERS1;
    softstate->ad_info.ad_defaults = &audiov686_defaults;
    bcopy(&audiov686_ch_cap, &softstate->ad_info.ad_play,
	  sizeof(ah_ad_ch_cap_t));
    bcopy(&audiov686_ch_cap, &softstate->ad_info.ad_record,
	  sizeof(ah_ad_ch_cap_t));
    softstate->ad_info.ad_play_comb = audiov686_cap_comb;
    softstate->ad_info.ad_rec_comb = audiov686_cap_comb;
    softstate->ad_info.ad_entry    = &audiov686_entry;
    softstate->ad_info.ad_dev_info = &softstate->dev_info;
    softstate->ad_info.ad_translate_flags =
				AD_MISC_8_P_TRANSLATE | AD_MISC_8_R_TRANSLATE;

    softstate->audio_handle = audio_hlp_attach(dip, cmd, &softstate->ad_info);
    if (!softstate->audio_handle) {
	cmn_err(CE_WARN, "failed to attach audio helper module");
	audiov686_cleanup(softstate);
	return DDI_FAILURE;
    }

    audio_hlp_set_private(softstate->audio_handle, softstate);


    /* map the audio controller's register block */
    status = ddi_regs_map_setup(dip, 1,
				(caddr_t *)&softstate->regs,
				0, V686_REGS_SIZE,
				&dev_attr, &softstate->regs_handle);

    if (status != DDI_SUCCESS) {
	if (status == DDI_REGS_ACC_CONFLICT) {
	    cmn_err(CE_WARN, "audio controller regs mapping conflicted with other mapping");
	} else {
	    cmn_err(CE_WARN, "audio controller regs mapping failed");
	}
	audiov686_cleanup(softstate);
	return DDI_FAILURE;
    }


    mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);

    if (ddi_get_iblock_cookie(dip, 0, &softstate->intr_cookie) != DDI_SUCCESS) {
	cmn_err(CE_WARN, "Could not get iblock cookie");
	audiov686_cleanup(softstate);
	return DDI_FAILURE;
    }
    mutex_init(&softstate->intr_mutex, NULL,
	       MUTEX_DRIVER, (void*)softstate->intr_cookie);

    if (ddi_add_intr(dip, 0,
		     &softstate->intr_cookie, NULL,
		     audiov686_intr, (caddr_t)softstate) != DDI_SUCCESS) {

	cmn_err(CE_WARN, "Could not add interrupt handler");
	audiov686_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * try to reset the audio controller, this verifies basic operation of the
     * sound hardware.
     */
    {
	static struct ac97_codec_ops codec_ops = {
	    audiov686_ac97_read,
	    audiov686_ac97_write
	};

	softstate->codec_handle = ac97_codec_init(softstate->audio_handle, &codec_ops, softstate);
    }

    if (audiov686_reset_card(softstate, &audiov686_defaults, TRUE) != AUDIOV686_SUCCESS) {
	cmn_err(CE_WARN, "reset of v686 audio controller failed");
	audiov686_cleanup(softstate);
	return DDI_FAILURE;	
    }


    if (audiov686_setup_dma(softstate) != AUDIOV686_SUCCESS) {
	cmn_err(CE_WARN, "dma setup of v686 audio controller failed");
	audiov686_cleanup(softstate);
	return DDI_FAILURE;	
    }

    if (audiov686_alloc_sample_buffer(softstate, &softstate->play_buf) != AUDIOV686_SUCCESS
	|| audiov686_alloc_sample_buffer(softstate, &softstate->record_buf) != AUDIOV686_SUCCESS) {
	cmn_err(CE_WARN, "cannot allocate sample buffers");
	audiov686_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * if we have an AC'97 codec without sample rate converter, 48kHz
     * is all we can play/record
     */
    if (!(softstate->ac97_ext_features & EAR_VRA)) {
	cmn_err(CE_NOTE, "no hardware sample rate converter available");
	bcopy(&audiov686_sample_rates_48k, &softstate->ad_info.ad_play.ad_srs,
	      sizeof(ah_ad_sample_rates_t));
	bcopy(&audiov686_sample_rates_48k, &softstate->ad_info.ad_record.ad_srs,
	      sizeof(ah_ad_sample_rates_t));
    }

    return DDI_SUCCESS;    
}


static int
audiov686_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audiov686_softstate *softstate = ddi_get_soft_state(audiov686_state, instance);

    DPRINTF4(AUDV686M_ATTACH, "detach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_DETACH:
	break;

    default:
	return DDI_FAILURE;
    }

    if (audiov686_cleanup(softstate) != DDI_SUCCESS) 
	return DDI_FAILURE;

    return DDI_SUCCESS;
}


/*
 * Audio Driver functions (for ah_ad_entry_t)
 */

static int    
audiov686_setup(audiohdl_t handle, int stream, int flags)
{
    DPRINTF4(AUDV686M_DRIVER, "audiov686_setup, stream %d, flags %x\n", stream, flags);

    return AUDIO_SUCCESS;
}


static void
audiov686_teardown(audiohdl_t handle, int stream, int dir)
{
    DPRINTF4(AUDV686M_DRIVER, "audiov686_teardown, stream %d, dir %d\n", stream, dir);
}


static int
audiov686_set_config(audiohdl_t handle, int stream,
		     int command, int dir, int arg1, int arg2)
{
    struct audiov686_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDV686M_DRIVER, "audiov686_set_config, stream %d, "
	     "command %d, dir %d, arg1 %d, arg2 %d\n",
	     stream, command, dir, arg1, arg2);

    CODEC_SET_CONFIG(softstate->codec_handle, command, dir, arg1, arg2);

    return AUDIO_SUCCESS;
}


static int
audiov686_set_format(audiohdl_t handle, int stream,
		 int dir, int sample_rate, int channels,
		 int precision, int encoding)
{
    struct audiov686_softstate *softstate = audio_hlp_get_private(handle);
    uint32_t type;

    DPRINTF4(AUDV686M_DRIVER, "audiov686_set_format, stream %d, dir %d, "
	     "sample_rate %d, channels %d, precision %d, encoding %d\n",
	     stream, dir, sample_rate, channels, precision, encoding);

    /* TODO: recording stuff */
#if 0
    if (dir == AUDIO_RECORD &&
	!(softstate->ac97_ext_features & EAR_VRA) && 
	sample_rate != AC97_SAMPLE_RATE_48000) {
	DPRINTF2(AUDV8233M_DRIVER, "unsupported sample rate %d\n", sample_rate);
	return AUDIO_FAILURE;
    }
#endif

    mutex_enter(&softstate->intr_mutex);

    audiov686_ac97_write(softstate, AC97_EXTENDED_AUDIO_STAT_CTRL_REGISTER, EASCR_VRA);
    if (dir == AUDIO_PLAY) {
	uint16_t val;

	DPRINTF2(AUDV686M_DRIVER, "set DAC VRA %d\n", sample_rate);
	audiov686_ac97_write(softstate, AC97_EXTENDED_FRONT_DAC_RATE_REGISTER, sample_rate);

	audiov686_ac97_read (softstate, AC97_EXTENDED_FRONT_DAC_RATE_REGISTER, &val);
	DPRINTF2(AUDV686M_DRIVER, "Front DAC samplerate is %d\n", val);
    }
    /* FIXME: recording stuff */
#if 0   
     else {
	DPRINTF4(AUDV8233M_DRIVER, "set ADC VRA %d\n", sample_rate);
	audiov8233_ac97_write(softstate, AC97_EXTENDED_LR_ADC_RATE_REGISTER, sample_rate);
	audiov8233_ac97_write(softstate, AC97_EXTENDED_MIC_ADC_RATE_REGISTER, sample_rate);
	
    } 
#endif
    if (dir == AUDIO_PLAY) {
	type = V686_GET8(V686_SGD_RDCHN_TYPE);

	if (precision == 16) 
	    type |= V686_SGD_TYPE_16BIT;
	else
	    type &= ~V686_SGD_TYPE_16BIT;

	if (channels == 2) 
	    type |= V686_SGD_TYPE_STEREO;
	else
	    type &= ~V686_SGD_TYPE_STEREO;

        /* I on Flag |  autoStart at eol */
        type |= V686_SGD_TYPE_AUTOSTART |  V686_SGD_TYPE_INTR_FLAG;
	V686_PUT8(V686_SGD_RDCHN_TYPE, type);
	    
	softstate->audio_info.play.sample_rate = sample_rate;
	softstate->audio_info.play.channels    = channels;
	softstate->audio_info.play.precision   = precision;
	softstate->audio_info.play.encoding    = encoding;
    }

#if 0
 else {
	V8233_PUT8(V8233_SGD_WRCHN_FIFO, V8233_SGD_FIFO_ENABLE);

	type = V8233_GET32(V8233_SGD_WRCHN_TYPE);

	if (precision == 16) 
	    type |= V8233_SGD_TYPE_16BIT;
	else
	    type &= ~V8233_SGD_TYPE_16BIT;

	if (channels == 2) 
	    type |= V8233_SGD_TYPE_STEREO;
	else
	    type &= ~V8233_SGD_TYPE_STEREO;


	V8233_PUT32(V8233_SGD_WRCHN_TYPE, type);
	/*printf("record chan type=0x%08x", V8233_GET32(V8233_SGD_WRCHN_TYPE));*/

	softstate->audio_info.record.sample_rate = sample_rate;
	softstate->audio_info.record.channels    = channels;
	softstate->audio_info.record.precision   = precision;
	softstate->audio_info.record.encoding    = encoding;
    }
#endif
    mutex_exit(&softstate->intr_mutex);

    return AUDIO_SUCCESS;
}

/*
 * This is called by timer created in xxx_fill_play_buffers().
 * NOTE: This function and all callee may not sleep!
 */
static void
audiov686_timed_fill_play_buf(void *arg)
{
    struct audiov686_softstate *softstate = arg;

    DPRINTF4(AUDV686M_DRIVER, "audiov686_timed_fill_play_buf\n");

    mutex_enter(&softstate->intr_mutex);
    softstate->play_timeout_id = 0;
    mutex_exit(&softstate->intr_mutex);

    audiov686_fill_play_buffers(softstate, 0);    
}

/*
 * Computes index to previous buffer (handles index underflow)
 */
#define PREV_INDEX(x) ( (x)>0 ? ((x)-1) : V686_NUM_SGD-1 )

/*
 * Fetch free sample buffers from the free list, fill them with audio
 * samples abtained from the audio framework and add them at the tail
 * of the VIA's SGD.
 * NOTE: It is called by audio library once only.
 * It creates timer and calls itself after a while through
 * audiov686_timed_fill_play_buf(), i. e., the data are polled. 
 * This functions retrieves data from audio library using
 * audiohlp_getaudio().
 * NOTE: this function may not sleep (that is cv_signal()/cv_wait() and
 * lot of other functions are not permitted).
 *
 */
static int
audiov686_fill_play_buffers(struct audiov686_softstate *softstate, int unpause)
{
    struct sgd_chunk *chunk;
    struct sample_buffer *buf = &softstate->play_buf;
    int bytes, len;
    int total_samples;
    int samples;
    int bindstat;
    int i;
    uint_t ccount;
    ddi_dma_cookie_t cookie;
    v686_buf_descr_t *bdesc;
    uint8_t ctrl;
    int byte_to_sample_shift;
    int end_of_playback = 0;
    int data_underrun = 0;

    total_samples = 0;

    byte_to_sample_shift = 0;
    if (softstate->audio_info.play.channels == 2) byte_to_sample_shift++;
    if (softstate->audio_info.play.precision == 16) byte_to_sample_shift++;

    mutex_enter(&softstate->intr_mutex);

    if (buf->io_started) { 
       /* try to free old buffers
	* NOTE: we may not free cur_idx, but also previous buffer yet.
	* It may be still in use by sound card.
	* Ommiting PREV_INDEX() test results in hard lockups very soon.
	*/
       while ( (buf->cur_idx != buf->head)
	     && (PREV_INDEX(buf->cur_idx) != buf->head) ){
	DPRINTF4(AUDV686M_INTR, "it: idx %d, head %d,tail %d\n",
	      buf->cur_idx,buf->head,buf->tail);
	  audiov686_release_play_buf_head(softstate, &softstate->play_buf);
       }
    } else {

	V686_PUT32 (V686_SGD_RDCHN_PTR, (uint32_t) (long) softstate->sgd_paddr->pcm_out_sgd);

	buf->head = 0;
	buf->tail = 0;
	buf->buffers_available = V686_NUM_SGD;
    }


    while ((chunk = buf->free) != NULL) {
	buf->free = chunk->next;
	chunk->next = NULL;

	bytes = SAMPLEBUF_INTR_CHUNK;
	samples = bytes >> byte_to_sample_shift;

	samples = audio_hlp_get_audio(softstate->audio_handle, chunk->data, 0, samples);
	DPRINTF4(AUDV686M_INTR, "got %d samples audio data\n", samples);
	if (samples <= 0) {
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}
	bytes = samples << byte_to_sample_shift;
	total_samples += samples;

	bindstat = ddi_dma_addr_bind_handle(chunk->dma_handle,
					    NULL,
					    chunk->data, chunk->reallen,
					    DDI_DMA_WRITE,
					    DDI_DMA_DONTWAIT, NULL,
					    &cookie, &ccount);
	if (bindstat != DDI_DMA_MAPPED) {
	    DPRINTF4(AUDV686M_INTR, "dma_addr_bind_handle failed %d\n", bindstat);
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}

	chunk->dma_bound = TRUE;
	buf->active[buf->tail] = chunk;

	for (i = 0; i < ccount && bytes > 0; i++) {
	    if (i > 0)
		ddi_dma_nextcookie(chunk->dma_handle, &cookie);
	    DPRINTF4(AUDV686M_INTR, "cookie #%d: at %p, %ld bytes\n",
		     i, (void*) (long) cookie.dmac_address, (long)cookie.dmac_size);

	    if (cookie.dmac_size < bytes)
		len = cookie.dmac_size;
	    else
		len = bytes;
	    bytes -= len;

	    buf->buffers_available--;

	    /*
	     * since we map DMA memory with the DDI_NEVERSWAP_ACC attribute, 
	     * we can access it without using ddi_{get,put}N
	     */
	    bdesc = softstate->sgd->pcm_out_sgd + buf->tail;
	    bdesc->bd_base = cookie.dmac_address;
	    bdesc->bd_len  = len;
	    bdesc->bd_stop = FALSE; 
	    bdesc->bd_flag = (bytes <= 0);
	    bdesc->bd_eol  = buf->tail == V686_NUM_SGD-1;

	    if (++buf->tail >= V686_NUM_SGD)
		buf->tail = 0;
	}
    }

    /*
     * If the sound card was paused (buffer underrun) and
     * if we have new data we will resume playback now
     */
       ctrl = V686_GET8(V686_SGD_RDCHN_CTRL);
       if (  ctrl & V686_SGD_CTRL_PAUSE) {
	  if ( unpause || total_samples > 0 ){
	     /* data & pause => resume playback */
	     ctrl &= ~V686_SGD_CTRL_PAUSE;
	     V686_PUT8(V686_SGD_RDCHN_CTRL, ctrl);
	  } else {
	     /* no data & pause => end of playback */
	     end_of_playback = 1;
	     data_underrun = (buf->free != NULL);
	  }
       }

    if ( total_samples > 0 || !buf->io_started ){
       if (!buf->io_started && buf->head != buf->tail) {
	   /* reprogram interrupt conditions */
	   uint8_t type = V686_GET8(V686_SGD_RDCHN_TYPE);
	   type |= V686_SGD_TYPE_AUTOSTART |  V686_SGD_TYPE_INTR_FLAG;
	   V686_PUT8(V686_SGD_RDCHN_TYPE, type);

	   /* start playback */
          ctrl = V686_GET8(V686_SGD_RDCHN_CTRL);
          ctrl &= ~V686_SGD_CTRL_PAUSE;
	  ctrl |= ( V686_SGD_CTRL_START  );
	  V686_PUT8(V686_SGD_RDCHN_CTRL, ctrl);
	     buf->io_started = TRUE;
       }
    }

    /* pull aditional data after a while... */
    if (!end_of_playback && !softstate->play_timeout_id) {
	/*
	 * free sample buffers left, and no more sample data available from
	 * the framework. Schedule a timeout routine to retry filling more
	 * buffers in time before the VIA AC has used all the samples we added
	 * to the SGD.
	 * 
	 * XXX The timeout (20ms) could be tuned a bit.  At 48kHz, 16bit, stereo,
	 * an 8K sample buffer plays ~41ms. At lower sample rates the timeout
	 * could reduced.
	 **/

        DPRINTF4(AUDV686M_INTR, "Entering timeout...");
	softstate->play_timeout_id = 
	    timeout(audiov686_timed_fill_play_buf, softstate,
		    drv_usectohz(20000));
    }


    mutex_exit(&softstate->intr_mutex);

    if (data_underrun)
	audio_hlp_play_shutdown(softstate->audio_handle, 0);

    DPRINTF4(AUDV686M_INTR, "DMA enabled, PCM out, status 0x%02x, ctrl 0x%02x, ptr 0x%08x, cnt 0x%08x\n",
	     V686_GET8(V686_SGD_RDCHN_STAT),
	     V686_GET8(V686_SGD_RDCHN_CTRL),
	     V686_GET32(V686_SGD_RDCHN_PTR),
	     V686_GET32(V686_SGD_RDCHN_CUR_CNT));

    return total_samples;
}



static int
audiov686_start_play(audiohdl_t handle, int stream)
{
    struct audiov686_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDV686M_DRIVER, "audiov686_start_play, stream %d\n", stream);
    
    audiov686_fill_play_buffers(softstate, 1);
    return AUDIO_SUCCESS;
}


static void
audiov686_pause_play(audiohdl_t handle, int stream)
{
    struct audiov686_softstate *softstate = audio_hlp_get_private(handle);
    uint8_t ctrl;

    DPRINTF4(AUDV686M_DRIVER, "audiov686_pause_play, stream %d\n", stream);

    mutex_enter(&softstate->intr_mutex);
    ctrl = V686_GET8(V686_SGD_RDCHN_CTRL);
    V686_PUT8(V686_SGD_RDCHN_CTRL, ctrl | V686_SGD_CTRL_PAUSE);
    mutex_exit(&softstate->intr_mutex);

}


static void
audiov686_stop_play(audiohdl_t handle, int stream)
{
    struct audiov686_softstate *softstate = audio_hlp_get_private(handle);
    int i;
    struct sgd_chunk *chunk;
    sample_buffer_t *buf = &softstate->play_buf;
    DPRINTF4(AUDV686M_DRIVER, "audiov686_stop_play, stream %d\n", stream);

    if (softstate->play_timeout_id) {
	untimeout(softstate->play_timeout_id);
	softstate->play_timeout_id = 0;
    }

    mutex_enter(&softstate->intr_mutex);
    softstate->play_buf.io_started = FALSE;

    V686_PUT8(V686_SGD_RDCHN_CTRL, V686_SGD_CTRL_TERMINATE);

    /*
     * throw away the DMA chunks that were active on the VIA "PCM in"
     * DMA engine, release them to the free list
     */
    for (i = 0; i < V686_NUM_SGD; i++) {
	if ((chunk = buf->active[i]) != NULL) {
	    buf->active[i] = NULL;

	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }
	    chunk->next = buf->free;
	    buf->free = chunk;
	}
    }

    mutex_exit(&softstate->intr_mutex);

}


static int
audiov686_start_record(audiohdl_t handle, int stream)
{
    /*struct audiov686_softstate *state = audio_hlp_get_private(handle);*/

    DPRINTF4(AUDV686M_DRIVER, "audiov686_start_record, stream %d\n", stream);

    return AUDIO_SUCCESS;
}


static void
audiov686_stop_record(audiohdl_t handle, int stream)
{
    /*struct audiov686_softstate *state = audio_hlp_get_private(handle);*/

    DPRINTF4(AUDV686M_DRIVER, "audiov686_stop_record, stream %d\n", stream);

}


static int
audiov686_ioctl(audiohdl_t handle, int channel,
		queue_t *q, mblk_t *mp, int *error)
{
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;

    DPRINTF4(AUDV686M_DRIVER, "audiov686_ioctl ioc_cmd=%x\n", iocp->ioc_cmd);

    switch (iocp->ioc_cmd) {
    default:
	return AD_NACK;
    }
}


static int
audiov686_iocdata(audiohdl_t handle, int channel,
		  queue_t *q, mblk_t *mp, int *error)
{
    struct copyresp *copyresp = (struct copyresp *)mp->b_rptr;
    mblk_t *data_mp;
    caddr_t arg = (caddr_t) copyresp->cp_private;

    if (mp->b_cont == NULL) return AD_NACK;
    data_mp = mp->b_cont;

    DPRINTF4(AUDV686M_DRIVER, "audiov686_iocdata cp_cmd=%x, arg=%p\n", copyresp->cp_cmd, arg);

    switch (copyresp->cp_cmd) {
    default:
	return AD_NACK;
    }
}


static audio_info_t audiov686_defaults = {
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_SPEAKER,		/* selected I/O port */
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_MICROPHONE,		/* selected I/O port */
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    AUDIO_MIN_GAIN,		/* monitor_gain */
    0,				/* output_muted */
#ifdef	AUDIOIO_HAS_REFCNT
    0,				/* driver reference count */
#endif
    {0, 0},			/* reserved fields */
#ifdef	AUDIOIO_HAS_FEATURES
    AUDIO_HWFEATURE_DUPLEX | AUDIO_HWFEATURE_IN2OUT | \
    	AUDIO_HWFEATURE_PLAY | AUDIO_HWFEATURE_RECORD,
    AUDIO_SWFEATURE_MIXER,
    0,
#endif
};


/*
 * Audio Driver ops vector definition
 */
static ah_ad_entry_t audiov686_entry = {
    audiov686_setup,
    audiov686_teardown,
    audiov686_set_config,
    audiov686_set_format,
    audiov686_start_play,
    audiov686_pause_play,
    audiov686_stop_play,
    audiov686_start_record,
    audiov686_stop_record,
    audiov686_ioctl,
    audiov686_iocdata,
};


static audio_device_t audiov686_dev_info = {
    "TOOLS,via686",		/* name */
    "1.0",			/* version */
    "via686"			/* config */
};


/* 0 term. list of sample rates (used as limits) */
static uint_t audiov686_srs[] = {
    8000, 48000, 0
};

/* 0 term. list of supported sample rates */
static uint_t audiov686_srs_48k[] = {
    AC97_SAMPLE_RATE_48000, 0
};

static ah_ad_sample_rates_t audiov686_sample_rates_48k = {
    AD_SRS_FLAG_SR_NOT_LIMITS, audiov686_srs_48k
};


static uint_t audiov686_chs[] = {
    AUDIO_CHANNELS_MONO,
    AUDIO_CHANNELS_STEREO,
    0,
};

static ah_ad_cap_comb_t audiov686_cap_comb[] = {
    { AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR },  /* 16-bit signed PCM */
    { AUDIO_PRECISION_8,  AUDIO_ENCODING_LINEAR8 }, /* 8-bit unsigned PCM */
    { 0 },
};


static ah_ad_ch_cap_t audiov686_ch_cap = {
    { AD_SRS_FLAG_SR_LIMITS, audiov686_srs },
    audiov686_chs,		/* list of channel types */
    0,				/* interrupt rate */
    2,				/* max channels */
    SAMPLEBUF_SIZE,		/* buffer size */
};
